import json

from unittest import mock
from urllib.parse import quote

import pytest

from snowflake.core import PollingOperation
from snowflake.core.procedure import (
    CallArgumentList,
    ColumnType,
    Procedure,
    ProcedureResource,
    ReturnDataType,
    ReturnTable,
    SQLFunction,
)
from snowflake.core.procedure._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
PROCEDURE = Procedure(
    name="my_proc",
    arguments=[],
    return_type=ReturnDataType(datatype="VARCHAR"),
    language_config=SQLFunction(),
    body="SELECT 'xyz'",
)


@pytest.fixture
def procedures(schema):
    return schema.procedures


@pytest.fixture
def procedure(procedures):
    return procedures["my_proc()"]


def test_create_procedure(fake_root, procedures):
    args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/procedures",
    )
    kwargs = extra_params(
        query_params=[],
        body={
            "name": "my_proc",
            "arguments": [],
            "return_type": {"datatype": "VARCHAR", "type": "DATATYPE"},
            "language_config": {"language": "SQL"},
            "body": "SELECT 'xyz'",
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        proc_res = procedures.create(PROCEDURE)
        assert isinstance(proc_res, ProcedureResource)
        assert proc_res.name_with_args == "my_proc()"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = procedures.create_async(PROCEDURE)
        assert isinstance(op, PollingOperation)
        proc_res = op.result()
        assert proc_res.name_with_args == "my_proc()"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_procedure(fake_root, procedures):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/procedures")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        it = procedures.iter()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = procedures.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_procedure(fake_root, procedure):
    from snowflake.core.procedure._generated.models import Procedure as ProcedureModel
    from snowflake.core.procedure._generated.models import ReturnDataType as ReturnDataTypeModel
    from snowflake.core.procedure._generated.models import SQLFunction as SQLFunctionModel

    model = ProcedureModel(
        name="my_proc",
        arguments=[],
        return_type=ReturnDataTypeModel(datatype="VARCHAR"),
        language_config=SQLFunctionModel(),
        body="SELECT 'xyz'",
    )
    args = (fake_root, "GET", BASE_URL + f"/databases/my_db/schemas/my_schema/procedures/{quote('my_proc()')}")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        my_proc = procedure.fetch()
        assert my_proc.to_dict() == PROCEDURE.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = procedure.fetch_async()
        assert isinstance(op, PollingOperation)
        my_proc = op.result()
        assert my_proc.to_dict() == PROCEDURE.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_procedure(fake_root, procedure):
    args = (
        fake_root,
        "DELETE",
        BASE_URL + f"/databases/my_db/schemas/my_schema/procedures/{quote('my_proc()')}",
    )
    kwargs = extra_params(query_params=[])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        procedure.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = procedure.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_call_procedure(fake_root, procedure):
    from snowflake.core.procedure._generated.models import Procedure as ProcedureModel
    from snowflake.core.procedure._generated.models import ReturnDataType as ReturnDataTypeModel
    from snowflake.core.procedure._generated.models import SQLFunction as SQLFunctionModel

    model = ProcedureModel(
        name="my_proc",
        arguments=[],
        return_type=ReturnDataTypeModel(datatype="VARCHAR"),
        language_config=SQLFunctionModel(),
        body="SELECT 'xyz'",
    )
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/procedures/my_proc:call")
    kwargs = extra_params(body={"call_arguments": []})
    fetch_response = mock_http_response(model.to_json())
    call_response = mock_http_response(json.dumps([{"result": "xyz"}]))

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.side_effect = [fetch_response, call_response]
        procedure.call(call_argument_list=CallArgumentList(call_arguments=[]))
    mocked_request.assert_called_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.side_effect = [fetch_response, call_response]
        op = procedure.call_async(call_argument_list=CallArgumentList(call_arguments=[]))
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_with(*args, **kwargs)


@pytest.mark.parametrize("extract, expected_result", [(True, {"a": 1, "b": 2}), (False, [{"foo": {"a": 1, "b": 2}}])])
@pytest.mark.parametrize("data_type", ["GEOMETRY", "GEOGRAPHY", "OBJECT", "VARIANT"])
def test_variant_mapping(procedure, extract, expected_result, fake_root, data_type):
    from snowflake.core.procedure._generated.models import Procedure as ProcedureModel
    from snowflake.core.procedure._generated.models import ReturnDataType as ReturnDataTypeModel
    from snowflake.core.procedure._generated.models import SQLFunction as SQLFunctionModel

    model = ProcedureModel(
        name="my_proc",
        arguments=[],
        return_type=ReturnDataTypeModel(datatype=data_type),
        language_config=SQLFunctionModel(),
        body="SELECT 'xyz'",
    )
    fetch_response = mock_http_response(model.to_json())
    call_response = mock_http_response(json.dumps([{"foo": '{\n  "a": 1,\n  "b": 2\n}'}]))

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.side_effect = [fetch_response, call_response]
        result = procedure.call(call_argument_list=CallArgumentList(call_arguments=[]), extract=extract)

    assert result == expected_result


@pytest.mark.parametrize("extract", [True, False])
def test_table_mapping(procedure, fake_root, extract):
    from snowflake.core.procedure._generated.models import Procedure as ProcedureModel
    from snowflake.core.procedure._generated.models import SQLFunction as SQLFunctionModel

    model = ProcedureModel(
        name="my_proc",
        arguments=[],
        return_type=ReturnTable(
            column_list=[ColumnType(name="id", datatype="NUMBER"), ColumnType(name="name", datatype="STRING")]
        ),
        language_config=SQLFunctionModel(),
        body="SELECT 'xyz'",
    )
    fetch_response = mock_http_response(model.to_json())
    call_response = mock_http_response(
        json.dumps([{"id": "1", "name": "jdoe"}, {"id": "2", "name": "jdoe"}, {"id": "3", "name": "jdoe"}])
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.side_effect = [fetch_response, call_response]
        result = procedure.call(call_argument_list=CallArgumentList(call_arguments=[]), extract=extract)

    assert result == [{"id": 1, "name": "jdoe"}, {"id": 2, "name": "jdoe"}, {"id": 3, "name": "jdoe"}]


def test_set_tags(fake_root, procedure, tag):
    args = (
        fake_root,
        "POST",
        BASE_URL + f"/databases/my_db/schemas/my_schema/procedures/{quote('my_proc()')}:set-tags",
    )
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        procedure.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = procedure.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, procedure, tag):
    args = (
        fake_root,
        "POST",
        BASE_URL + f"/databases/my_db/schemas/my_schema/procedures/{quote('my_proc()')}:unset-tags",
    )
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        procedure.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = procedure.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, procedure):
    args = (
        fake_root,
        "GET",
        BASE_URL + f"/databases/my_db/schemas/my_schema/procedures/{quote('my_proc()')}:get-tags",
    )
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert procedure.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = procedure.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



"""Manages Snowflake dynamic table."""

from ..dynamic_table._generated.models import (
    DownstreamLag,
    DynamicTable,
    DynamicTableClone,
    DynamicTableColumn,
    ErrorResponse,
    PointOfTime,
    PointOfTimeOffset,
    PointOfTimeStatement,
    PointOfTimeTimestamp,
    SuccessAcceptedResponse,
    SuccessResponse,
    TargetLag,
    UserDefinedLag,
)
from ._dynamic_table import DynamicTableCollection, DynamicTableResource


# TODO: validate if all classes should be exported
__all__ = [
    "DynamicTableResource",
    "DynamicTableCollection",
    "DownstreamLag",
    "DynamicTable",
    "DynamicTableClone",
    "DynamicTableColumn",
    "ErrorResponse",
    "PointOfTime",
    "PointOfTimeOffset",
    "PointOfTimeStatement",
    "PointOfTimeTimestamp",
    "SuccessAcceptedResponse",
    "SuccessResponse",
    "TargetLag",
    "UserDefinedLag",
]



"""
Snowflake Account API.

The Snowflake Account API is a REST API that you can use to access, update, and perform certain actions on Account resource in Snowflake.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

import warnings

from typing import TYPE_CHECKING, Iterator, Optional, TypeVar

from snowflake.core._common import AccountObjectCollectionParent, ObjectReferenceMixin
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._operation import PollingOperation, PollingOperations
from snowflake.core._utils import get_function_name_with_args, replace_function_name_in_name_with_args
from snowflake.core.account._generated.models.account import Account

from .account_api import AccountApi


if TYPE_CHECKING:
    from snowflake.core import Root


T = TypeVar("T")


class AccountCollectionBase(AccountObjectCollectionParent["AccountResource"]):
    _identifier_requires_args: bool = False

    def __init__(self, root: "Root", resource_class: type[T]) -> None:
        super().__init__(root, resource_class)
        self._api = AccountApi(
            root=self.root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @api_telemetry
    def create(self, account: Account) -> "AccountResource":  # noqa: F821
        """Create an account.

        Parameters
        __________
        account: Account
             (required)
        """
        self._api.create_account(
            account=account,
            async_req=False,
        )
        if self._identifier_requires_args:
            return self._ref_class(get_function_name_with_args(account), self)
        return self._ref_class(account.name, self)

    @api_telemetry
    def create_async(self, account: Account) -> PollingOperation["AccountResource"]:  # noqa: F821
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self._api.create_account(
            account=account,
            async_req=True,
        )
        if self._identifier_requires_args:
            return PollingOperation(future, lambda _: self._ref_class(get_function_name_with_args(account), self))
        return PollingOperation(
            future, lambda _: self._ref_class(account if isinstance(account, str) else account.name, self)
        )

    @api_telemetry
    def iter(
        self,
        *,
        like: Optional[str] = None,
        limit: Optional[int] = None,
        history: Optional[bool] = None,
        show_limit: Optional[int] = None,
    ) -> Iterator[Account]:
        """List accounts.

        Parameters
        __________
        like: str
             Parameter to filter the command output by resource name. Uses case-insensitive pattern matching, with support for SQL wildcard characters.
        limit: int
             Parameter to limit the maximum number of rows returned by a command.
        history: bool
             Optionally includes dropped accounts that have not yet been purged.
        """
        if show_limit is not None:
            if limit is not None:
                raise ValueError("Cannot specify both 'limit' and 'show_limit'. Use only one of them.")
            warnings.warn(DeprecationWarning("'show_limit' is deprecated, use 'limit' instead."))  # noqa: B028
            limit = show_limit

        resources = self._api.list_accounts(
            like=like,
            show_limit=limit,
            history=history,
            async_req=False,
        )
        return iter(resources)

    @api_telemetry
    def iter_async(
        self,
        *,
        like: Optional[str] = None,
        limit: Optional[int] = None,
        history: Optional[bool] = None,
        show_limit: Optional[int] = None,
    ) -> PollingOperation[Iterator[Account]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        if show_limit is not None:
            if limit is not None:
                raise ValueError("Cannot specify both 'limit' and 'show_limit'. Use only one of them.")
            warnings.warn(DeprecationWarning("'show_limit' is deprecated, use 'limit' instead."))  # noqa: B028
            limit = show_limit

        future = self._api.list_accounts(
            like=like,
            show_limit=limit,
            history=history,
            async_req=True,
        )

        return PollingOperations.iterator(future)


class AccountResourceBase(ObjectReferenceMixin["AccountCollection"]):
    _collection_class: AccountCollectionBase
    _identifier_requires_args: bool = False
    _plural_name: str

    def __init__(self, name: str, collection_class: AccountCollectionBase) -> None:
        self.collection = collection_class
        if self._identifier_requires_args:
            self.name_with_args = name
        else:
            self.name = name

    @api_telemetry
    def undrop(
        self,
    ) -> None:
        """Restore an account.

        Parameters
        __________
        """
        self.collection._api.undrop_account(
            self._identifier,
            async_req=False,
        )

    @api_telemetry
    def undrop_async(
        self,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`undrop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.undrop_account(
            self._identifier,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def drop(
        self,
        grace_period_in_days: int,
        if_exists: Optional[bool] = None,
    ) -> None:
        """Delete an account.

        Parameters
        __________
        grace_period_in_days: int
             Specifies the number of days during which the account can be restored (“undropped”). The minimum is 3 days and the maximum is 90 days. (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`: The endpoint does not throw an error if the resource does not exist. It returns a 200 success response, but does not take any action on the resource. - `false`: The endpoint throws an error if the resource doesn't exist.
        """
        self.collection._api.delete_account(
            self._identifier,
            grace_period_in_days=grace_period_in_days,
            if_exists=if_exists,
            async_req=False,
        )

    @api_telemetry
    def drop_async(
        self,
        grace_period_in_days: int,
        if_exists: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.delete_account(
            self._identifier,
            grace_period_in_days=grace_period_in_days,
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @property
    def _identifier(self) -> str:
        """Get the identifier for this resource."""
        if self._identifier_requires_args:
            return self.name_with_args
        return self.name

    def _set_new_name(self, target_name: str) -> None:
        """Set a new name for the resource."""
        if self._identifier_requires_args:
            old_name = self.name_with_args
            self.name_with_args = replace_function_name_in_name_with_args(self.name_with_args, target_name)
        else:
            old_name = self.name
            self.name = target_name

        # Update the collection reference to point to the new name
        self.collection.update_reference(old_name=old_name, new_name=self._identifier, resource=self)



from typing import TYPE_CHECKING

from ._generated.api.artifact_repository_api_base import (
    ArtifactRepositoryCollectionBase,
    ArtifactRepositoryResourceBase,
)


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class ArtifactRepositoryCollection(ArtifactRepositoryCollectionBase):
    """Represents the collection operations on the Snowflake artifact repository resource.

    With this collection, you can create, iterate through, and search for artifact repository that you have access to
    in the current context.
    """

    def __init__(self, schema: "SchemaResource") -> None:
        super().__init__(schema, ArtifactRepositoryResource)


class ArtifactRepositoryResource(ArtifactRepositoryResourceBase):
    """Represents a reference to a Snowflake artifact repository.

    With this artifact repository reference, you can fetch information about an artifact repository,
    as well as perform certain actions on it.
    """

    _plural_name = "artifact_repositories"



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#


import pytest

from urllib3 import HTTPResponse

from snowflake.core._root import Root
from snowflake.core.rest import SSEClient
from snowflake.core.table import Table, TableColumn


TEST_SERVICE_NAME = "SNOWPY_TEST_SERVICE"
TEST_TABLE_NAME = "SNOWPY_TEST_TABLE"

API_NOT_ENABLED = "Cortex Agent API is not enabled"

pytestmark = [pytest.mark.skip_gov]


@pytest.fixture(autouse=True)
def setup_cortex_search_service(connection, warehouse, tables, backup_warehouse_fixture):
    del backup_warehouse_fixture
    # Base Table
    tables.create(
        Table(
            name=TEST_TABLE_NAME,
            columns=[
                TableColumn(name="search_col", datatype="varchar"),
                TableColumn(name="filter1", datatype="varchar"),
            ],
        ),
        mode="ifnotexists",
    )

    with connection.cursor() as cursor:
        cursor.execute(f"use warehouse {warehouse.name}").fetchone()

        rows = ",".join(
            [
                "('Citi reimagines financial transactions in the data cloud', 'filter_value_1')",
                "('ADP works in the data cloud. This is another mention of data cloud.', 'filter_value_2')",
                "('Lorem ipsum dolor sit amet', 'filter_value_1')",
                "('Cortex Search is a search solution for any problem', 'filter_value_2')",
            ]
        )
        cursor.execute(f"INSERT INTO {TEST_TABLE_NAME} VALUES {rows}")

        # Cortex Search Service
        cursor.execute(
            f"CREATE OR REPLACE CORTEX SEARCH SERVICE {TEST_SERVICE_NAME} "
            f"ON search_col TARGET_LAG='1 minute' WAREHOUSE={warehouse.name} "
            f"AS SELECT search_col, filter1 FROM {TEST_TABLE_NAME}"
        )

        try:
            yield
        finally:
            cursor.execute(f"DROP CORTEX SEARCH SERVICE {TEST_SERVICE_NAME}")
            cursor.execute(f"DROP TABLE {TEST_TABLE_NAME}")


@pytest.fixture(autouse=True)
def precheck_cortex_search_enabled(cortex_search_services, setup_cortex_search_service):
    del setup_cortex_search_service
    try:
        cortex_search_services[TEST_SERVICE_NAME].search("hi", ["search_col", "filter1"], limit=5)
    except Exception as err:
        if API_NOT_ENABLED in err.body:
            pytest.xfail(API_NOT_ENABLED)
        raise


# Test agent_run through CortexAgentService
# this is just to test the integrity of the python -> cortex agent_run service.
# We would not try to match the exact response as the response might change over time.
def test_agent_run(root: Root, database, schema):
    resp = root.cortex_agent_service.Run(
        {
            "model": "mistral-7b",
            "messages": [{"role": "user", "content": [{"type": "text", "text": "I have an internet issues"}]}],
            "tools": [{"tool_spec": {"type": "cortex_search", "name": "search1"}}],
            "tool_resources": {"search1": {"name": f"{database.name}.{schema.name}.{TEST_SERVICE_NAME}"}},
        }
    )

    for e in resp.events():
        # check if the event and data are not None
        assert e.event is not None
        assert e.data is not None


# Test client parsing of agent_run service response through XP endpoint
def test_agent_run_xp(root: Root):
    def validate_text_message(event):
        content = event.data["delta"]["content"][0]
        assert content["index"] == 1
        assert content["type"] == "text"
        assert content["text"] == "Let me analyze your data."

    def validate_tool_use(event):
        content = event.data["delta"]["content"][0]
        assert content["index"] == 2
        assert content["type"] == "tool_use"
        tool_use = content["tool_use"]
        assert tool_use["tool_use_id"] == "tool_123"
        assert tool_use["name"] == "Analyst1"
        assert tool_use["input"]["sql_query"] == "SELECT * FROM sales"

    def validate_tool_result(event):
        content = event.data["delta"]["content"][0]
        assert content["index"] == 3
        assert content["type"] == "tool_results"
        tool_results = content["tool_results"]
        assert tool_results["tool_use_id"] == "tool_123"
        assert tool_results["status"] == "success"
        assert tool_results["content"][0]["type"] == "text"
        assert tool_results["content"][0]["text"] == "Query executed successfully"

    def validate_tool_result_json(event):
        content = event.data["delta"]["content"][0]
        assert content["index"] == 3
        assert content["type"] == "tool_results"
        tool_results = content["tool_results"]
        assert tool_results["tool_use_id"] == "tool_123"
        assert tool_results["status"] == "success"
        assert tool_results["content"][0]["type"] == "json"
        assert tool_results["content"][0]["json"]["sql_query"] == "SELECT * FROM sales"

    def validate_error_response(event):
        assert event.data["code"] == "400"
        assert event.data["message"] == "Invalid tool configuration"

    def validate_error_event(event):
        assert event.data["code"] == "399505"
        assert event.data["message"] == "Internal server error"

    test_cases = [
        # Text Message case
        {
            "body": (
                '[ {"event":"message.delta","data":{"id":"REQUEST_ID","object":"message.delta",'
                '"delta":{"content":[{"index":1,"type":"text","text":"Let me analyze your data."}]}}},'
                '{"event":"done","data":"[DONE]"}]'
            ),
            "validator": validate_text_message,
        },
        # Tool Use case
        {
            "body": (
                '[ {"event":"message.delta","data":{"id":"REQUEST_ID","object":"message.delta",'
                '"delta":{"content":[{"index":2,"type":"tool_use","tool_use":{"tool_use_id":"tool_123",'
                '"name":"Analyst1","input":{"sql_query":"SELECT * FROM sales"}}}]}}},'
                '{"event":"done","data":"[DONE]"}]'
            ),
            "validator": validate_tool_use,
        },
        # Tool Result case
        {
            "body": (
                '[ {"event":"message.delta","data":{"id":"REQUEST_ID","object":"message.delta",'
                '"delta":{"content":[{"index":3,"type":"tool_results","tool_results":{"tool_use_id":"tool_123",'
                '"content":[{"type":"text","text":"Query executed successfully"}],"status":"success"}}]}}},'
                '{"event":"done","data":"[DONE]"}]'
            ),
            "validator": validate_tool_result,
        },
        # Tool Result with JSON case
        {
            "body": (
                '[ {"event":"message.delta","data":{"id":"REQUEST_ID","object":"message.delta",'
                '"delta":{"content":[{"index":3,"type":"tool_results","tool_results":{"tool_use_id":"tool_123",'
                '"content":[{"type":"json","json":{"sql_query":"SELECT * FROM sales"}}],"status":"success"}}]}}},'
                '{"event":"done","data":"[DONE]"}]'
            ),
            "validator": validate_tool_result_json,
        },
        # Error Response case
        {
            "body": (
                '[ {"event":"error","data":{"code":"400","message":"Invalid tool configuration"}},'
                '{"event":"done","data":"[DONE]"}]'
            ),
            "validator": validate_error_response,
        },
        # Error Event case
        {
            "body": (
                '[ {"event":"error","data":{"code":"399505","message":"Internal server error"}},'
                '{"event":"done","data":"[DONE]"}]'
            ),
            "validator": validate_error_event,
        },
    ]

    for test_case in test_cases:
        response = HTTPResponse(
            body=test_case["body"].encode("utf-8"),
            status=200,
            headers={"Content-Type": "application/json"},
            preload_content=False,
        )
        res = SSEClient(response)

        event_num = 0
        for e in res.events():
            if event_num == 0:
                assert e.event in ["message.delta", "error"]
                assert e.data is not None

                if e.event == "message.delta":
                    assert e.data["id"] == "REQUEST_ID"
                    assert e.data["object"] == "message.delta"
                    test_case["validator"](e)
                elif e.event == "error":
                    test_case["validator"](e)

            elif event_num == 1:
                assert e.event == "done"
                assert e.data == "[DONE]"

            event_num += 1

        assert event_num == 2  # Ensure we got both the main event and the done event



"""
Snowflake Stage API.

The Snowflake Stage API is a REST API that you can use to access, update, and perform certain actions on stage resources in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictBool, StrictStr


class StageDirectoryTable(BaseModel):
    """A model object representing the StageDirectoryTable resource.

    Constructs an object of type StageDirectoryTable with the provided properties.

    Parameters
    __________
    enable : bool,  default False
        Specifies whether to add a directory table to the stage. When the value is TRUE, a directory table is created with the stage.
    refresh_on_create : bool,  default True
        Specifies whether to automatically refresh the directory table metadata once, immediately after the stage is created.
    auto_refresh : bool,  default False
        Specifies whether Snowflake should enable triggering automatic refreshes of the directory table metadata when new or updated data files are available in the named external stage specified in the URL value.
    notification_integration : str, optional
        Specifies the name of the notification integration used to automatically refresh the directory table metadata.
    """

    enable: Optional[StrictBool] = False

    refresh_on_create: Optional[StrictBool] = True

    auto_refresh: Optional[StrictBool] = False

    notification_integration: Optional[StrictStr] = None

    __properties = ["enable", "refresh_on_create", "auto_refresh", "notification_integration"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> StageDirectoryTable:
        """Create an instance of StageDirectoryTable from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> StageDirectoryTable:
        """Create an instance of StageDirectoryTable from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return StageDirectoryTable.model_validate(obj)

        _obj = StageDirectoryTable.model_validate(
            {
                "enable": obj.get("enable") if obj.get("enable") is not None else False,
                "refresh_on_create": obj.get("refresh_on_create") if obj.get("refresh_on_create") is not None else True,
                "auto_refresh": obj.get("auto_refresh") if obj.get("auto_refresh") is not None else False,
                "notification_integration": obj.get("notification_integration"),
            }
        )

        return _obj


class StageDirectoryTableModel:
    def __init__(
        self,  # optional properties
        enable: Optional[bool] = False,
        refresh_on_create: Optional[bool] = True,
        auto_refresh: Optional[bool] = False,
        notification_integration: Optional[str] = None,
    ):
        """A model object representing the StageDirectoryTable resource.

        Constructs an object of type StageDirectoryTable with the provided properties.

        Parameters
        __________
        enable : bool,  default False
            Specifies whether to add a directory table to the stage. When the value is TRUE, a directory table is created with the stage.
        refresh_on_create : bool,  default True
            Specifies whether to automatically refresh the directory table metadata once, immediately after the stage is created.
        auto_refresh : bool,  default False
            Specifies whether Snowflake should enable triggering automatic refreshes of the directory table metadata when new or updated data files are available in the named external stage specified in the URL value.
        notification_integration : str, optional
            Specifies the name of the notification integration used to automatically refresh the directory table metadata.
        """
        self.enable = enable
        self.refresh_on_create = refresh_on_create
        self.auto_refresh = auto_refresh
        self.notification_integration = notification_integration

    __properties = ["enable", "refresh_on_create", "auto_refresh", "notification_integration"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return StageDirectoryTable(
            enable=self.enable,
            refresh_on_create=self.refresh_on_create,
            auto_refresh=self.auto_refresh,
            notification_integration=self.notification_integration,
        )

    @classmethod
    def _from_model(cls, model) -> StageDirectoryTableModel:
        return StageDirectoryTableModel(
            enable=model.enable,
            refresh_on_create=model.refresh_on_create,
            auto_refresh=model.auto_refresh,
            notification_integration=model.notification_integration,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a StageDirectoryTable.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the StageDirectoryTable object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> StageDirectoryTableModel:
        """Create an instance of StageDirectoryTable from a dict.

        This method constructs a StageDirectoryTable object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        StageDirectoryTable
            A StageDirectoryTable object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(StageDirectoryTable.from_dict(obj))


StageDirectoryTable._model_class = StageDirectoryTableModel



import time

from contextlib import suppress

import pytest

from snowflake.core.alert import Alert, MinutesSchedule
from snowflake.core.exceptions import NotFoundError
from snowflake.core.table import Table, TableColumn
from tests.utils import random_string


@pytest.mark.min_sf_ver("8.36.0")
def test_execute(alerts, tables):
    alert_name = random_string(10, "test_alert_")
    temp_table = tables.create(
        Table(
            name=random_string(10, "test_table_"),
            columns=[TableColumn(name="c1", datatype="varchar"), TableColumn(name="c2", datatype="varchar")],
        )
    )

    alert_handle = alerts.create(
        Alert(
            name=alert_name,
            condition="SELECT 1",
            action=f"DROP TABLE {temp_table.name}",
            schedule=MinutesSchedule(minutes=1),
            comment="ThIs iS a ComMeNT",
        )
    )

    try:
        temp_table.fetch()

        alert_handle.execute()

        # Wait for the alert to execute
        time.sleep(61)

        with pytest.raises(NotFoundError):
            temp_table.fetch()
    finally:
        with suppress(NotFoundError):
            temp_table.drop()
        alert_handle.drop()



"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict

from snowflake.core.cortex.analyst_service._generated.models.tableau_server_files import (
    TableauServerFiles,
    TableauServerFilesModel,
)
from snowflake.core.cortex.analyst_service._generated.models.tableau_stage_files import (
    TableauStageFiles,
    TableauStageFilesModel,
)


class TableauInput(BaseModel):
    """A model object representing the TableauInput resource.

    Constructs an object of type TableauInput with the provided properties.

    Parameters
    __________
    stage_files : TableauStageFiles, optional

    server_files : TableauServerFiles, optional
    """

    stage_files: Optional[TableauStageFiles] = None

    server_files: Optional[TableauServerFiles] = None

    __properties = ["stage_files", "server_files"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> TableauInput:
        """Create an instance of TableauInput from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of stage_files
        if self.stage_files:
            _dict["stage_files"] = self.stage_files.to_dict()

        # override the default output from pydantic by calling `to_dict()` of server_files
        if self.server_files:
            _dict["server_files"] = self.server_files.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> TableauInput:
        """Create an instance of TableauInput from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return TableauInput.model_validate(obj)

        _obj = TableauInput.model_validate(
            {
                "stage_files": TableauStageFiles.from_dict(obj.get("stage_files"))
                if obj.get("stage_files") is not None
                else None,
                "server_files": TableauServerFiles.from_dict(obj.get("server_files"))
                if obj.get("server_files") is not None
                else None,
            }
        )

        return _obj


class TableauInputModel:
    def __init__(
        self,  # optional properties
        stage_files: Optional[TableauStageFiles] = None,
        server_files: Optional[TableauServerFiles] = None,
    ):
        """A model object representing the TableauInput resource.

        Constructs an object of type TableauInput with the provided properties.

        Parameters
        __________
        stage_files : TableauStageFiles, optional

        server_files : TableauServerFiles, optional
        """
        self.stage_files = stage_files
        self.server_files = server_files

    __properties = ["stage_files", "server_files"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return TableauInput(
            stage_files=self.stage_files._to_model() if self.stage_files is not None else None,
            server_files=self.server_files._to_model() if self.server_files is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> TableauInputModel:
        return TableauInputModel(
            stage_files=TableauStageFilesModel._from_model(model.stage_files) if model.stage_files else None,
            server_files=TableauServerFilesModel._from_model(model.server_files) if model.server_files else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a TableauInput.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the TableauInput object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> TableauInputModel:
        """Create an instance of TableauInput from a dict.

        This method constructs a TableauInput object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        TableauInput
            A TableauInput object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(TableauInput.from_dict(obj))


TableauInput._model_class = TableauInputModel



"""
Snowflake Alert API.

The Snowflake Alert API is a REST API that you can use to access, update, and perform certain actions on Alert resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictStr

from snowflake.core.alert._generated.models.point_of_time import PointOfTime, PointOfTimeModel


class AlertClone(BaseModel):
    """A model object representing the AlertClone resource.

    Constructs an object of type AlertClone with the provided properties.

    Parameters
    __________
    name : str
        Name of the alert
    point_of_time : PointOfTime, optional
    """

    name: StrictStr

    point_of_time: Optional[PointOfTime] = None

    __properties = ["name", "point_of_time"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> AlertClone:
        """Create an instance of AlertClone from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of point_of_time
        if self.point_of_time:
            _dict["point_of_time"] = self.point_of_time.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> AlertClone:
        """Create an instance of AlertClone from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return AlertClone.model_validate(obj)

        _obj = AlertClone.model_validate(
            {
                "name": obj.get("name"),
                "point_of_time": PointOfTime.from_dict(obj.get("point_of_time"))
                if obj.get("point_of_time") is not None
                else None,
            }
        )

        return _obj


class AlertCloneModel:
    def __init__(
        self,
        name: str,
        # optional properties
        point_of_time: Optional[PointOfTime] = None,
    ):
        """A model object representing the AlertClone resource.

        Constructs an object of type AlertClone with the provided properties.

        Parameters
        __________
        name : str
            Name of the alert
        point_of_time : PointOfTime, optional
        """
        self.name = name
        self.point_of_time = point_of_time

    __properties = ["name", "point_of_time"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return AlertClone(
            name=self.name,
            point_of_time=self.point_of_time._to_model() if self.point_of_time is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> AlertCloneModel:
        return AlertCloneModel(
            name=model.name,
            point_of_time=PointOfTimeModel._from_model(model.point_of_time) if model.point_of_time else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a AlertClone.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the AlertClone object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> AlertCloneModel:
        """Create an instance of AlertClone from a dict.

        This method constructs a AlertClone object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        AlertClone
            A AlertClone object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(AlertClone.from_dict(obj))


AlertClone._model_class = AlertCloneModel



import random
import string
import time

from collections.abc import Sequence
from typing import Any
from unittest.mock import MagicMock

import pytest

from snowflake.core._utils import check_version_gte
from snowflake.core.version import __version__ as VERSION


BASE_URL = "http://localhost:80/api/v2"


def is_gov_deployment(snowflake_region: str) -> bool:
    return "GOV" in snowflake_region


def is_preprod_deployment(snowflake_region: str) -> bool:
    return "PREPROD" in snowflake_region


def is_prod_deployment(version_str: str) -> bool:
    return version_str and all(character.isdigit() or character == "." for character in version_str)


def is_prod_or_preprod(version_str: str, snowflake_region: str) -> bool:
    # For prod, we check if version string is all digits or decimals, because non-prod versions contain
    # letters or other symbols.
    # For preprod, we check if "PREPROD" is in the regions name
    return is_prod_deployment(version_str) or is_preprod_deployment(snowflake_region)


def ensure_snowflake_version(current_version: str, requested_version: str) -> None:
    version_number = extract_version_number(current_version)
    if not check_version_gte(version_number, requested_version):
        pytest.skip(
            f"Skipping test because the current server version {version_number} "
            f"is older than the minimum version {requested_version}"
        )


def extract_version_number(version_str: str) -> str:
    # on non-prod deployments the CURRENT_VERSION result contains the version number and hash
    return version_str.split(" ")[0]


def random_string(
    length: int, prefix: str = "", suffix: str = "", choices: Sequence[str] = string.ascii_lowercase
) -> str:
    """Our convenience function to generate random string for object names.

    Args:
        length: How many random characters to choose from choices.
            length would be at least 6 for avoiding collision
        prefix: Prefix to add to random string generated.
        suffix: Suffix to add to random string generated.
        choices: A generator of things to choose from.
    """
    random_part = "".join([random.choice(choices) for _ in range(length)]) + str(time.time_ns())

    return "".join([prefix, random_part, suffix])


def unquote(name: str) -> str:
    if '""' in name:
        return name.replace('""', '"')
    if name.startswith('"') and name.endswith('"'):
        return name[1:-1]


def extra_params(**kwargs) -> dict[str, Any]:
    headers = {"Accept": "application/json", "User-Agent": "python_api/" + VERSION}
    if "body" in kwargs:
        headers["Content-Type"] = "application/json"
    return {
        "query_params": [],
        "headers": headers,
        "post_params": [],
        "body": None,
        "_preload_content": True,
        "_request_timeout": None,
    } | kwargs


def mock_http_response(data: str = "") -> MagicMock:
    m = MagicMock()
    m.data = data
    m.status = 200
    return m



from unittest import mock

from snowflake.core import PollingOperation
from snowflake.core.database import Database, DatabaseResource

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


def test_create_database(fake_root, dbs):
    database = Database(name="sophie_db", kind="TRANSIENT", comment="This is Sophie's database", trace_level="always")
    args = (fake_root, "POST", BASE_URL + "/databases?createMode=errorIfExists")
    kwargs = extra_params(
        query_params=[("createMode", "errorIfExists")],
        body={
            "name": "sophie_db",
            "kind": "TRANSIENT",
            "comment": "This is Sophie's database",
            "trace_level": "always",
            "dropped_on": None,
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db_res = dbs.create(database)
        assert isinstance(db_res, DatabaseResource)
        assert db_res.name == database.name
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = dbs.create_async(database)
        assert isinstance(op, PollingOperation)
        db_res = op.result()
        assert db_res.name == database.name
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_database(fake_root, dbs):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = dbs.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(fake_root, "GET", BASE_URL + "/databases", **extra_params())


def test_fetch_database(fake_root, db):
    from snowflake.core.database._generated.models import Database as DatabaseModel

    args = (fake_root, "GET", BASE_URL + "/databases/my_db")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(DatabaseModel(name="my_db").to_json())
        database = db.fetch()
        assert database.to_dict() == Database("my_db").to_dict()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(DatabaseModel(name="my_db").to_json())
        op = db.fetch_async()
        assert isinstance(op, PollingOperation)
        database = op.result()
        assert database.to_dict() == Database("my_db").to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_database(fake_root, db):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_undrop_database(fake_root, db):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db:undrop")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.undrop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.undrop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_enable_replication(fake_root, db):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/replication:enable?ignore_edition_check=False")
    kwargs = extra_params(query_params=[("ignore_edition_check", False)], body={"accounts": ["acc"]})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.enable_replication(["acc"])
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.enable_replication_async(["acc"])
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_disable_replication(fake_root, db):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/replication:disable")
    kwargs = extra_params(body={"accounts": []})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.disable_replication()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.disable_replication_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_refresh_replication(fake_root, db):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/replication:refresh")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.refresh_replication()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.refresh_replication_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_enable_failover(fake_root, db):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/failover:enable")
    kwargs = extra_params(body={"accounts": ["acc"]})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.enable_failover(["acc"])
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.enable_failover_async(["acc"])
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_disable_failover(fake_root, db):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/failover:disable")
    kwargs = extra_params(body={"accounts": []})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.disable_failover()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.disable_failover_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_promote_to_primary_failover(fake_root, db):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/failover:primary")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        db.promote_to_primary_failover()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = db.promote_to_primary_failover_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)



# coding: utf-8

# flake8: noqa

"""
Snowflake Managed Account API.

The Snowflake Managed Account API is a REST API that you can use to access, update, and perform certain actions on Managed Account resource in Snowflake.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from snowflake.core.managed_account._generated.api.managed_account_api import ManagedAccountApi

# import models into sdk package
from snowflake.core.managed_account._generated.models.error_response import ErrorResponse
from snowflake.core.managed_account._generated.models.managed_account import ManagedAccount
from snowflake.core.managed_account._generated.models.success_accepted_response import SuccessAcceptedResponse
from snowflake.core.managed_account._generated.models.success_response import SuccessResponse



import uuid

import snowflake.connector

from ..utils import (
    backup_database_and_schema,
    backup_role,
    backup_warehouse,
    connection_config,
    connection_keys,
    create_and_use_new_database_and_schema,
    create_zip_from_paths,
    upload_given_files_to_stage,
)


TEST_DATABASE_NAME = "DATABASE_PYTHON_TESTING_STOREDPROC"
TEST_SCHEMA = "GH_JOB_{}".format(str(uuid.uuid4()).replace("-", "_"))
STAGE_NAME = "STAGE_PYTHON_TEST_STOREDPROC"


storedproc_create_sql = """
CREATE OR REPLACE PROCEDURE testing_storedproc()
RETURNS STRING
LANGUAGE PYTHON
RUNTIME_VERSION = '3.13'
PACKAGES = ('snowflake-snowpark-python', 'asn1crypto','atpublic','certifi','cffi','charset-normalizer','cryptography','docker-py','filelock','idna','packaging','platformdirs','pycparser','pydantic','pyjwt','pyopenssl','pytest', 'pytest-xdist','pytest-cov','python-dateutil','urllib3','tomlkit','sortedcontainers')
HANDLER = 'run_integration_test'
execute as caller
AS
$$
def run_integration_test(session):
    import sys
    import traceback

    output_file_path = "/tmp/execution_output.txt"
    snowlfake_core_zip_file_name = "snowflake_core.zip"
    stage_name = "STAGE_PYTHON_TEST_STOREDPROC"

    with open(output_file_path, "w") as f:
        sys.stdout = f
        sys.stderr = f
        try:
            import zipfile
            import os
            import pytest
            import pytest_cov

            os.environ["RUNNING_IN_STOREDPROC"] = "True"
            os.environ["PYTHONUNBUFFERED"] = "1"
            should_enable_logging = (os.environ.get("ENABLE_DEBUG_LOGGING_IN_SNOWPY") == "True")

            def clear_modules(prefix):
                for module_name in list(sys.modules.keys()):
                    if module_name.startswith(prefix):
                        del sys.modules[module_name]

            clear_modules("snowflake.core")

            get_status = session.file.get(f"@{stage_name}/zip/{snowlfake_core_zip_file_name}", "/tmp/zip")
            if len(get_status) != 1:
                raise Exception("not able to load the snowflake_core")

            with zipfile.ZipFile(f"/tmp/zip/{snowlfake_core_zip_file_name}", "r") as zip_ref:
                zip_ref.extractall("/tmp/expanded/snowflake_core")
            sys.path.insert(0, "/tmp/expanded/snowflake_core/src")
            print("Successfully completed the set up for executing the integration test in storedproc")
            try:
                if should_enable_logging:
                    import logging
                    import pathlib
                    import snowflake.core
                    snowflake.core.simple_file_logging(
                        path=pathlib.Path("execution_output.txt"),
                        level=logging.DEBUG,
                        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                    )
                exit_code = pytest.main(
                    ["/tmp/expanded/snowflake_core/tests/integ", "-W", "ignore", "-r", "a", "--no-cov", "-v", "-m", "not flaky"]
                )
                if exit_code != 0:
                    print("Some tests failed")
                else:
                    print("All tests passed")
                print("Successfully executed integration test in storedproc")
            except Exception as e:
                traceback.print_exc(file=f)
                print("Failed to execute integration test in storedproc")
        except Exception as e:
            traceback.print_exc(file=f)
            print("Failed to complete the set up for executing the integration test in storedproc")

    put_result = session.file.put(output_file_path, f"@{stage_name}", auto_compress=False)
$$;

"""


def run():
    execution_output = None
    config = connection_config()

    # copy the snowpy package to the tests/integ/storedproc folder
    snowflake_core_source_folder = "./src"
    snowflake_core_test_folder = "./tests"
    snowflake_core_pyproject_file_path = "./pyproject.toml"
    storedproc_directory_path = "tests/integ/storedproc"
    snowflake_core_zip_filename = "snowflake_core.zip"
    storedproc_execution_output_file_name = "execution_output.txt"
    snowflake_core_zip_filepath = f"{storedproc_directory_path}/{snowflake_core_zip_filename}"

    # create the zip file
    create_zip_from_paths(
        [snowflake_core_source_folder, snowflake_core_test_folder, snowflake_core_pyproject_file_path],
        snowflake_core_zip_filepath,
    )

    _keys = connection_keys()

    with snowflake.connector.connect(**{k: config[k] for k in _keys if k in config}) as connection:
        cursor = connection.cursor()
        with backup_database_and_schema(cursor), backup_warehouse(cursor), backup_role(cursor):
            try:
                create_and_use_new_database_and_schema(cursor, TEST_DATABASE_NAME, TEST_SCHEMA)

                # create the stage and put the files in the stage
                cursor.execute(f"CREATE STAGE {STAGE_NAME}")
                upload_given_files_to_stage(cursor, f"{STAGE_NAME}/zip", [snowflake_core_zip_filepath])

                # try to create and execute the storedproc
                try:
                    cursor.execute(storedproc_create_sql)
                    cursor.execute("call testing_storedproc()")
                except Exception as e:
                    print(f"Error creating and executing the storedproc: {e.with_traceback(None)}")

                # fetch the output of the storedproc execution
                try:
                    # need to specify full Stage name if tests mess up the current database and schema
                    cursor.execute(
                        f"GET @{TEST_DATABASE_NAME}.{TEST_SCHEMA}.{STAGE_NAME}/{storedproc_execution_output_file_name} file://{storedproc_directory_path}/"
                    )
                    storedproc_execution_output_file_path = (
                        f"{storedproc_directory_path}/{storedproc_execution_output_file_name}"
                    )
                    with open(f"{storedproc_execution_output_file_path}") as f:
                        execution_output = f.read()
                except Exception as e:
                    raise Exception(f"Error reading the storedproc output: {e.with_traceback(None)}") from e

                # read the execution output
                print("Storedproc Output::::::STARTS")
                print(execution_output)
                if not (execution_output.find("All tests passed") != -1):
                    raise Exception("Some Integration test are failing")
                print("Storedproc Output::::::ENDS")
            finally:
                cursor.execute(f"USE DATABASE {TEST_DATABASE_NAME}")
                cursor.execute(f"USE SCHEMA {TEST_SCHEMA}")
                cursor.execute("DROP PROCEDURE IF EXISTS testing_storedproc()")
                cursor.execute(f"DROP STAGE  IF EXISTS {STAGE_NAME}")
                cursor.execute(f"DROP SCHEMA IF EXISTS {TEST_SCHEMA}")


if __name__ == "__main__":
    run()



"""
Snowflake Compute Pools API.

The Snowflake Compute Pools API is a REST API that you can use to access, update, and perform common actions on Compute Pool resources.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated


class ComputePool(BaseModel):
    """A model object representing the ComputePool resource.

    Constructs an object of type ComputePool with the provided properties.

    Parameters
    __________
    name : str
        A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
    min_nodes : int
        Minimum number of nodes for the compute pool.
    max_nodes : int
        Maximum number of nodes for the compute pool.
    instance_family : str
        Instance family for the compute pool. Use the /api/v2/compute-pools/instance-families endpoint to retrieve available values.
    auto_resume : bool, optional
        Whether Snowflake automatically resumes the compute pool when any statement that requires the compute pool is submitted.
    comment : str, optional
        Comment describing the compute pool.
    state : str, optional
        Current state of the compute pool. Possible values include UNKNOWN, STARTING, IDLE, ACTIVE, STOPPING, SUSPENDED, and RESIZING — **Read-only:** *any user-provided value will be ignored.*
    num_services : int, optional
        Number of services on the compute pool — **Read-only:** *any user-provided value will be ignored.*
    num_jobs : int, optional
        Number of jobs on the compute pool — **Read-only:** *any user-provided value will be ignored.*
    auto_suspend_secs : int, optional
        Number of seconds until the compute pool automatically suspends.
    active_nodes : int, optional
        Number of currently active nodes on the compute pool — **Read-only:** *any user-provided value will be ignored.*
    idle_nodes : int, optional
        Number of currently idle nodes on the compute pool — **Read-only:** *any user-provided value will be ignored.*
    target_nodes : int, optional
        Number of target nodes on the compute pool — **Read-only:** *any user-provided value will be ignored.*
    created_on : datetime, optional
        Time the compute pool was created — **Read-only:** *any user-provided value will be ignored.*
    resumed_on : datetime, optional
        Time the compute pool was last resumed — **Read-only:** *any user-provided value will be ignored.*
    updated_on : datetime, optional
        Time the compute pool was last updated — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Identifier for the current owner of the compute pool — **Read-only:** *any user-provided value will be ignored.*
    is_exclusive : bool, optional
        Whether a compute pool is created exclusively for a Snowflake Native App — **Read-only:** *any user-provided value will be ignored.*
    application : str, optional
        Name of the Snowflake Native App if the compute pool is created exclusively for the app — **Read-only:** *any user-provided value will be ignored.*
    budget : str, optional
        The name of the budget monitoring the credit usage of the compute pool — **Read-only:** *any user-provided value will be ignored.*
    error_code : str, optional
        Current error the compute pool hit if any — **Read-only:** *any user-provided value will be ignored.*
    status_message : str, optional
        Current status of the compute pool if any — **Read-only:** *any user-provided value will be ignored.*
    desired_nodes : int, optional
        Number of nodes to maintain on the compute pool — **Read-only:** *any user-provided value will be ignored.*
    desired_nodes_expire_after : datetime, optional
        Time after which the desired nodes on the compute pool will expire — **Read-only:** *any user-provided value will be ignored.*
    optimize_for_capacity : bool, optional
        Whether the compute pool is optimized for capacity — **Read-only:** *any user-provided value will be ignored.*
    placement_group : str, optional
        The name of the placement group on the compute pool — **Read-only:** *any user-provided value will be ignored.*
    """

    name: Annotated[str, Field(strict=True)]

    min_nodes: StrictInt

    max_nodes: StrictInt

    instance_family: StrictStr

    auto_resume: Optional[StrictBool] = None

    comment: Optional[StrictStr] = None

    state: Optional[StrictStr] = None

    num_services: Optional[StrictInt] = None

    num_jobs: Optional[StrictInt] = None

    auto_suspend_secs: Optional[StrictInt] = None

    active_nodes: Optional[StrictInt] = None

    idle_nodes: Optional[StrictInt] = None

    target_nodes: Optional[StrictInt] = None

    created_on: Optional[datetime] = None

    resumed_on: Optional[datetime] = None

    updated_on: Optional[datetime] = None

    owner: Optional[StrictStr] = None

    is_exclusive: Optional[StrictBool] = None

    application: Optional[StrictStr] = None

    budget: Optional[StrictStr] = None

    error_code: Optional[StrictStr] = None

    status_message: Optional[StrictStr] = None

    desired_nodes: Optional[StrictInt] = None

    desired_nodes_expire_after: Optional[datetime] = None

    optimize_for_capacity: Optional[StrictBool] = None

    placement_group: Optional[StrictStr] = None

    __properties = [
        "name",
        "min_nodes",
        "max_nodes",
        "instance_family",
        "auto_resume",
        "comment",
        "state",
        "num_services",
        "num_jobs",
        "auto_suspend_secs",
        "active_nodes",
        "idle_nodes",
        "target_nodes",
        "created_on",
        "resumed_on",
        "updated_on",
        "owner",
        "is_exclusive",
        "application",
        "budget",
        "error_code",
        "status_message",
        "desired_nodes",
        "desired_nodes_expire_after",
        "optimize_for_capacity",
        "placement_group",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("state")
    def state_validate_enum(cls, v):
        if v is None:
            return v
        if v not in ("UNKNOWN", "STARTING", "IDLE", "ACTIVE", "STOPPING", "SUSPENDED", "RESIZING"):
            raise ValueError(
                "must validate the enum values ('UNKNOWN','STARTING','IDLE','ACTIVE','STOPPING','SUSPENDED','RESIZING')"
            )
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ComputePool:
        """Create an instance of ComputePool from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "state",
                    "num_services",
                    "num_jobs",
                    "active_nodes",
                    "idle_nodes",
                    "target_nodes",
                    "created_on",
                    "resumed_on",
                    "updated_on",
                    "owner",
                    "is_exclusive",
                    "application",
                    "budget",
                    "error_code",
                    "status_message",
                    "desired_nodes",
                    "desired_nodes_expire_after",
                    "optimize_for_capacity",
                    "placement_group",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> ComputePool:
        """Create an instance of ComputePool from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return ComputePool.model_validate(obj)

        _obj = ComputePool.model_validate(
            {
                "name": obj.get("name"),
                "min_nodes": obj.get("min_nodes"),
                "max_nodes": obj.get("max_nodes"),
                "instance_family": obj.get("instance_family"),
                "auto_resume": obj.get("auto_resume"),
                "comment": obj.get("comment"),
                "state": obj.get("state"),
                "num_services": obj.get("num_services"),
                "num_jobs": obj.get("num_jobs"),
                "auto_suspend_secs": obj.get("auto_suspend_secs"),
                "active_nodes": obj.get("active_nodes"),
                "idle_nodes": obj.get("idle_nodes"),
                "target_nodes": obj.get("target_nodes"),
                "created_on": obj.get("created_on"),
                "resumed_on": obj.get("resumed_on"),
                "updated_on": obj.get("updated_on"),
                "owner": obj.get("owner"),
                "is_exclusive": obj.get("is_exclusive"),
                "application": obj.get("application"),
                "budget": obj.get("budget"),
                "error_code": obj.get("error_code"),
                "status_message": obj.get("status_message"),
                "desired_nodes": obj.get("desired_nodes"),
                "desired_nodes_expire_after": obj.get("desired_nodes_expire_after"),
                "optimize_for_capacity": obj.get("optimize_for_capacity"),
                "placement_group": obj.get("placement_group"),
            }
        )

        return _obj


class ComputePoolModel:
    def __init__(
        self,
        name: str,
        min_nodes: int,
        max_nodes: int,
        instance_family: str,
        # optional properties
        auto_resume: Optional[bool] = None,
        comment: Optional[str] = None,
        state: Optional[str] = None,
        num_services: Optional[int] = None,
        num_jobs: Optional[int] = None,
        auto_suspend_secs: Optional[int] = None,
        active_nodes: Optional[int] = None,
        idle_nodes: Optional[int] = None,
        target_nodes: Optional[int] = None,
        created_on: Optional[datetime] = None,
        resumed_on: Optional[datetime] = None,
        updated_on: Optional[datetime] = None,
        owner: Optional[str] = None,
        is_exclusive: Optional[bool] = None,
        application: Optional[str] = None,
        budget: Optional[str] = None,
        error_code: Optional[str] = None,
        status_message: Optional[str] = None,
        desired_nodes: Optional[int] = None,
        desired_nodes_expire_after: Optional[datetime] = None,
        optimize_for_capacity: Optional[bool] = None,
        placement_group: Optional[str] = None,
    ):
        """A model object representing the ComputePool resource.

        Constructs an object of type ComputePool with the provided properties.

        Parameters
        __________
        name : str
            A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
        min_nodes : int
            Minimum number of nodes for the compute pool.
        max_nodes : int
            Maximum number of nodes for the compute pool.
        instance_family : str
            Instance family for the compute pool. Use the /api/v2/compute-pools/instance-families endpoint to retrieve available values.
        auto_resume : bool, optional
            Whether Snowflake automatically resumes the compute pool when any statement that requires the compute pool is submitted.
        comment : str, optional
            Comment describing the compute pool.
        state : str, optional
            Current state of the compute pool. Possible values include UNKNOWN, STARTING, IDLE, ACTIVE, STOPPING, SUSPENDED, and RESIZING.
        num_services : int, optional
            Number of services on the compute pool.
        num_jobs : int, optional
            Number of jobs on the compute pool.
        auto_suspend_secs : int, optional
            Number of seconds until the compute pool automatically suspends.
        active_nodes : int, optional
            Number of currently active nodes on the compute pool.
        idle_nodes : int, optional
            Number of currently idle nodes on the compute pool.
        target_nodes : int, optional
            Number of target nodes on the compute pool.
        created_on : datetime, optional
            Time the compute pool was created.
        resumed_on : datetime, optional
            Time the compute pool was last resumed.
        updated_on : datetime, optional
            Time the compute pool was last updated.
        owner : str, optional
            Identifier for the current owner of the compute pool.
        is_exclusive : bool, optional
            Whether a compute pool is created exclusively for a Snowflake Native App.
        application : str, optional
            Name of the Snowflake Native App if the compute pool is created exclusively for the app.
        budget : str, optional
            The name of the budget monitoring the credit usage of the compute pool.
        error_code : str, optional
            Current error the compute pool hit if any.
        status_message : str, optional
            Current status of the compute pool if any.
        desired_nodes : int, optional
            Number of nodes to maintain on the compute pool
        desired_nodes_expire_after : datetime, optional
            Time after which the desired nodes on the compute pool will expire.
        optimize_for_capacity : bool, optional
            Whether the compute pool is optimized for capacity.
        placement_group : str, optional
            The name of the placement group on the compute pool.
        """
        self.name = name
        self.min_nodes = min_nodes
        self.max_nodes = max_nodes
        self.instance_family = instance_family
        self.auto_resume = auto_resume
        self.comment = comment
        self.state = state
        self.num_services = num_services
        self.num_jobs = num_jobs
        self.auto_suspend_secs = auto_suspend_secs
        self.active_nodes = active_nodes
        self.idle_nodes = idle_nodes
        self.target_nodes = target_nodes
        self.created_on = created_on
        self.resumed_on = resumed_on
        self.updated_on = updated_on
        self.owner = owner
        self.is_exclusive = is_exclusive
        self.application = application
        self.budget = budget
        self.error_code = error_code
        self.status_message = status_message
        self.desired_nodes = desired_nodes
        self.desired_nodes_expire_after = desired_nodes_expire_after
        self.optimize_for_capacity = optimize_for_capacity
        self.placement_group = placement_group

    __properties = [
        "name",
        "min_nodes",
        "max_nodes",
        "instance_family",
        "auto_resume",
        "comment",
        "state",
        "num_services",
        "num_jobs",
        "auto_suspend_secs",
        "active_nodes",
        "idle_nodes",
        "target_nodes",
        "created_on",
        "resumed_on",
        "updated_on",
        "owner",
        "is_exclusive",
        "application",
        "budget",
        "error_code",
        "status_message",
        "desired_nodes",
        "desired_nodes_expire_after",
        "optimize_for_capacity",
        "placement_group",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return ComputePool(
            name=self.name,
            min_nodes=self.min_nodes,
            max_nodes=self.max_nodes,
            instance_family=self.instance_family,
            auto_resume=self.auto_resume,
            comment=self.comment,
            state=self.state,
            num_services=self.num_services,
            num_jobs=self.num_jobs,
            auto_suspend_secs=self.auto_suspend_secs,
            active_nodes=self.active_nodes,
            idle_nodes=self.idle_nodes,
            target_nodes=self.target_nodes,
            created_on=self.created_on,
            resumed_on=self.resumed_on,
            updated_on=self.updated_on,
            owner=self.owner,
            is_exclusive=self.is_exclusive,
            application=self.application,
            budget=self.budget,
            error_code=self.error_code,
            status_message=self.status_message,
            desired_nodes=self.desired_nodes,
            desired_nodes_expire_after=self.desired_nodes_expire_after,
            optimize_for_capacity=self.optimize_for_capacity,
            placement_group=self.placement_group,
        )

    @classmethod
    def _from_model(cls, model) -> ComputePoolModel:
        return ComputePoolModel(
            name=model.name,
            min_nodes=model.min_nodes,
            max_nodes=model.max_nodes,
            instance_family=model.instance_family,
            auto_resume=model.auto_resume,
            comment=model.comment,
            state=model.state,
            num_services=model.num_services,
            num_jobs=model.num_jobs,
            auto_suspend_secs=model.auto_suspend_secs,
            active_nodes=model.active_nodes,
            idle_nodes=model.idle_nodes,
            target_nodes=model.target_nodes,
            created_on=model.created_on,
            resumed_on=model.resumed_on,
            updated_on=model.updated_on,
            owner=model.owner,
            is_exclusive=model.is_exclusive,
            application=model.application,
            budget=model.budget,
            error_code=model.error_code,
            status_message=model.status_message,
            desired_nodes=model.desired_nodes,
            desired_nodes_expire_after=model.desired_nodes_expire_after,
            optimize_for_capacity=model.optimize_for_capacity,
            placement_group=model.placement_group,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a ComputePool.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the ComputePool object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> ComputePoolModel:
        """Create an instance of ComputePool from a dict.

        This method constructs a ComputePool object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        ComputePool
            A ComputePool object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(ComputePool.from_dict(obj))


ComputePool._model_class = ComputePoolModel



"""
Snowflake User Defined Function API.

The Snowflake User Defined Function API is a REST API that you can use to access, update, and perform certain actions on User Defined Function resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import ConfigDict, StrictBool, StrictStr

from snowflake.core.user_defined_function._generated.models.return_type import ReturnType


class ReturnDataType(ReturnType):
    """A model object representing the ReturnDataType resource.

    Constructs an object of type ReturnDataType with the provided properties.

    Parameters
    __________
    datatype : str
        Return data type
    nullable : bool, optional
        Argument null return acceptance criteria
    """

    datatype: StrictStr

    nullable: Optional[StrictBool] = None

    __properties = ["type"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ReturnDataType:
        """Create an instance of ReturnDataType from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        _dict["type"] = ReturnType.get_child_model_discriminator_value("ReturnDataType")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> ReturnDataType:
        """Create an instance of ReturnDataType from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return ReturnDataType.model_validate(obj)

        _obj = ReturnDataType.model_validate(
            {
                "datatype": obj.get("datatype"),
                "nullable": obj.get("nullable"),
            }
        )

        return _obj


class ReturnDataTypeModel(ReturnType):
    def __init__(
        self,
        datatype: str,
        # optional properties
        nullable: Optional[bool] = None,
    ):
        """A model object representing the ReturnDataType resource.

        Constructs an object of type ReturnDataType with the provided properties.

        Parameters
        __________
        datatype : str
            Return data type
        nullable : bool, optional
            Argument null return acceptance criteria
        """
        super().__init__()
        self.datatype = datatype
        self.nullable = nullable

    __properties = ["type"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return ReturnDataType(
            datatype=self.datatype,
            nullable=self.nullable,
        )

    @classmethod
    def _from_model(cls, model) -> ReturnDataTypeModel:
        return ReturnDataTypeModel(
            datatype=model.datatype,
            nullable=model.nullable,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a ReturnDataType.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the ReturnDataType object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> ReturnDataTypeModel:
        """Create an instance of ReturnDataType from a dict.

        This method constructs a ReturnDataType object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        ReturnDataType
            A ReturnDataType object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(ReturnDataType.from_dict(obj))


ReturnDataType._model_class = ReturnDataTypeModel



# coding: utf-8

# flake8: noqa

"""
Cortex Chat API.

OpenAPI 3.0 specification for the Cortex Chat API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from snowflake.core.cortex.chat_service._generated.api.cortex_chat_api import CortexChatApi

# import models into sdk package
from snowflake.core.cortex.chat_service._generated.models.chat_request import ChatRequest
from snowflake.core.cortex.chat_service._generated.models.chat_request_messages_inner import ChatRequestMessagesInner
from snowflake.core.cortex.chat_service._generated.models.chat_request_search_services_inner import (
    ChatRequestSearchServicesInner,
)
from snowflake.core.cortex.chat_service._generated.models.chat_request_semantic_models_inner import (
    ChatRequestSemanticModelsInner,
)
from snowflake.core.cortex.chat_service._generated.models.error_response import ErrorResponse



import pytest

from snowflake.core.grant._grant import Grant
from snowflake.core.grant._grantee import Grantees
from snowflake.core.grant._privileges import Privileges
from snowflake.core.grant._securables import Securables


@pytest.mark.use_accountadmin
@pytest.mark.usefixtures("backup_warehouse_fixture")
def test_list_grant(test_role_name, test_function_name, test_warehouse_name, grants, session, schema):
    # grant roles

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.function(f"{test_function_name}(number, number)"),
            privileges=[Privileges.all_privileges],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.future_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select, Privileges.insert],
        )
    )

    # check roles have been granted
    grants_to_role = session.sql(f"SHOW GRANTS TO ROLE {test_role_name}").collect()
    grants_list = [grant for grant in grants_to_role]
    assert len(grants_list) > 0
    grants_to_role_list_with_api = [grants for grants in grants.to(Grantees.role(test_role_name))]
    assert len(grants_to_role_list_with_api) == len(grants_list)



import pytest


pytestmark = pytest.mark.usefixtures("backup_database_schema")


@pytest.mark.min_sf_ver("9.8.0")
def test_fetch(temp_schema, temp_schema_case_sensitive):
    schema = temp_schema.fetch()
    assert schema.name.upper() == temp_schema.name.upper()

    schema = temp_schema_case_sensitive.fetch()
    assert f'"{schema.name}"' == temp_schema_case_sensitive.name
    assert schema.comment == "created by temp_schema_case_sensitive"
    assert schema.serverless_task_min_statement_size == "XSMALL"
    assert schema.serverless_task_max_statement_size == "X2LARGE"
    assert schema.user_task_managed_initial_warehouse_size == "MEDIUM"
    assert schema.suspend_task_after_num_failures == 10
    assert schema.user_task_timeout_ms == 3600000



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import pytest

from snowflake.core._common import CreateMode
from snowflake.core.exceptions import APIError, ConflictError
from snowflake.core.service import Service, ServiceSpecStageFile
from tests.utils import random_string


pytestmark = [pytest.mark.skip_gov]


def test_create(services, service_spec_file_on_stage, shared_compute_pool):
    try:
        service_name = random_string(5, "test_service_")

        test_service = Service(
            name=service_name,
            compute_pool=shared_compute_pool,
            spec=service_spec_file_on_stage,
            min_instances=1,
            max_instances=1,
        )
        s = services.create(test_service)

        fetched_service = s.fetch()
        assert fetched_service.name == service_name.upper()

        # already existing service
        with pytest.raises(ConflictError):
            services.create(test_service)

        s = services.create(test_service, mode=CreateMode.if_not_exists)

        s.drop()

        # case sensitive name for service is not supported
        service_name = random_string(5, "test_service_")
        service_name = f'"{service_name}"'
        test_service = Service(
            name=service_name,
            compute_pool=shared_compute_pool,
            spec=service_spec_file_on_stage,
            min_instances=1,
            max_instances=1,
        )

        with pytest.raises(APIError):
            services.create(test_service)
    finally:
        services[service_name].drop(if_exists=True)


def test_create_or_replace(services, session, imagerepo, shared_compute_pool):
    service_name = random_string(5, "test_service_")
    stage_name = random_string(5, "test_stage_")
    spec_file = "spec.yaml"
    test_service = Service(
        name=service_name,
        compute_pool=shared_compute_pool,
        spec=ServiceSpecStageFile(stage=stage_name, spec_file=spec_file),
        min_instances=1,
        max_instances=1,
    )
    with pytest.raises(ValueError):
        services.create(test_service, mode=CreateMode.or_replace)


def test_create_with_spec_inline(services, temp_service_from_spec_inline, database):
    service = services[temp_service_from_spec_inline.name].fetch()
    assert (
        service.name == temp_service_from_spec_inline.name  # for mixed case names
        or service.name.upper() == temp_service_from_spec_inline.name.upper()  # for upper/lower case names
    )
    assert 'name: "hello-world"' in service.spec.spec_text
    assert service.comment == "created by temp_service_from_spec_inline"



import pytest

from snowflake.core.alert import Alert, MinutesSchedule
from snowflake.core.exceptions import NotFoundError
from tests.utils import random_string


@pytest.mark.min_sf_ver("8.36.0")
def test_drop(alerts):
    alert_name = random_string(10, "test_alert_")

    alert_handle = alerts.create(
        Alert(
            name=alert_name,
            condition="SELECT 1",
            action="SELECT 2",
            schedule=MinutesSchedule(minutes=1),
            comment="ThIs iS a ComMeNT",
        )
    )

    alert_handle.drop()

    with pytest.raises(NotFoundError):
        alert_handle.drop()

    alert_handle.drop(if_exists=True)



"""
Snowflake Services API.

The Snowflake Services API is a REST API that you can use to access, update, and perform certain actions on Services resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated


class ServiceContainer(BaseModel):
    """A model object representing the ServiceContainer resource.

    Constructs an object of type ServiceContainer with the provided properties.

    Parameters
    __________
    database_name : str, optional
        A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
    schema_name : str, optional
        A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
    service_name : str, optional
        A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
    service_status : str, optional
        The current status of the service — **Read-only:** *any user-provided value will be ignored.*
    instance_id : str, optional
        ID of the service instance (this is the index of the service instance starting from 0) — **Read-only:** *any user-provided value will be ignored.*
    instance_status : str, optional
        The current status of the service instance — **Read-only:** *any user-provided value will be ignored.*
    container_name : str, optional
        Name of the container — **Read-only:** *any user-provided value will be ignored.*
    status : str, optional
        Service container status — **Read-only:** *any user-provided value will be ignored.*
    message : str, optional
        Additional clarification about status — **Read-only:** *any user-provided value will be ignored.*
    image_name : str, optional
        Image name used to create the service container — **Read-only:** *any user-provided value will be ignored.*
    image_digest : str, optional
        The unique and immutable identifier representing the image content — **Read-only:** *any user-provided value will be ignored.*
    restart_count : int, optional
        Number of times Snowflake restarted the service — **Read-only:** *any user-provided value will be ignored.*
    start_time : str, optional
        Date and time when the container started — **Read-only:** *any user-provided value will be ignored.*
    """

    database_name: Optional[Annotated[str, Field(strict=True)]] = None

    schema_name: Optional[Annotated[str, Field(strict=True)]] = None

    service_name: Optional[Annotated[str, Field(strict=True)]] = None

    service_status: Optional[StrictStr] = None

    instance_id: Optional[StrictStr] = None

    instance_status: Optional[StrictStr] = None

    container_name: Optional[StrictStr] = None

    status: Optional[StrictStr] = None

    message: Optional[StrictStr] = None

    image_name: Optional[StrictStr] = None

    image_digest: Optional[StrictStr] = None

    restart_count: Optional[StrictInt] = None

    start_time: Optional[StrictStr] = None

    __properties = [
        "database_name",
        "schema_name",
        "service_name",
        "service_status",
        "instance_id",
        "instance_status",
        "container_name",
        "status",
        "message",
        "image_name",
        "image_digest",
        "restart_count",
        "start_time",
    ]

    @field_validator("database_name")
    def database_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("schema_name")
    def schema_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("service_name")
    def service_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ServiceContainer:
        """Create an instance of ServiceContainer from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "service_status",
                    "instance_id",
                    "instance_status",
                    "container_name",
                    "status",
                    "message",
                    "image_name",
                    "image_digest",
                    "restart_count",
                    "start_time",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> ServiceContainer:
        """Create an instance of ServiceContainer from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return ServiceContainer.model_validate(obj)

        _obj = ServiceContainer.model_validate(
            {
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "service_name": obj.get("service_name"),
                "service_status": obj.get("service_status"),
                "instance_id": obj.get("instance_id"),
                "instance_status": obj.get("instance_status"),
                "container_name": obj.get("container_name"),
                "status": obj.get("status"),
                "message": obj.get("message"),
                "image_name": obj.get("image_name"),
                "image_digest": obj.get("image_digest"),
                "restart_count": obj.get("restart_count"),
                "start_time": obj.get("start_time"),
            }
        )

        return _obj


class ServiceContainerModel:
    def __init__(
        self,  # optional properties
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        service_name: Optional[str] = None,
        service_status: Optional[str] = None,
        instance_id: Optional[str] = None,
        instance_status: Optional[str] = None,
        container_name: Optional[str] = None,
        status: Optional[str] = None,
        message: Optional[str] = None,
        image_name: Optional[str] = None,
        image_digest: Optional[str] = None,
        restart_count: Optional[int] = None,
        start_time: Optional[str] = None,
    ):
        """A model object representing the ServiceContainer resource.

        Constructs an object of type ServiceContainer with the provided properties.

        Parameters
        __________
        database_name : str, optional
            A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
        schema_name : str, optional
            A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
        service_name : str, optional
            A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
        service_status : str, optional
            The current status of the service.
        instance_id : str, optional
            ID of the service instance (this is the index of the service instance starting from 0).
        instance_status : str, optional
            The current status of the service instance.
        container_name : str, optional
            Name of the container.
        status : str, optional
            Service container status.
        message : str, optional
            Additional clarification about status.
        image_name : str, optional
            Image name used to create the service container.
        image_digest : str, optional
            The unique and immutable identifier representing the image content.
        restart_count : int, optional
            Number of times Snowflake restarted the service.
        start_time : str, optional
            Date and time when the container started.
        """
        self.database_name = database_name
        self.schema_name = schema_name
        self.service_name = service_name
        self.service_status = service_status
        self.instance_id = instance_id
        self.instance_status = instance_status
        self.container_name = container_name
        self.status = status
        self.message = message
        self.image_name = image_name
        self.image_digest = image_digest
        self.restart_count = restart_count
        self.start_time = start_time

    __properties = [
        "database_name",
        "schema_name",
        "service_name",
        "service_status",
        "instance_id",
        "instance_status",
        "container_name",
        "status",
        "message",
        "image_name",
        "image_digest",
        "restart_count",
        "start_time",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return ServiceContainer(
            database_name=self.database_name,
            schema_name=self.schema_name,
            service_name=self.service_name,
            service_status=self.service_status,
            instance_id=self.instance_id,
            instance_status=self.instance_status,
            container_name=self.container_name,
            status=self.status,
            message=self.message,
            image_name=self.image_name,
            image_digest=self.image_digest,
            restart_count=self.restart_count,
            start_time=self.start_time,
        )

    @classmethod
    def _from_model(cls, model) -> ServiceContainerModel:
        return ServiceContainerModel(
            database_name=model.database_name,
            schema_name=model.schema_name,
            service_name=model.service_name,
            service_status=model.service_status,
            instance_id=model.instance_id,
            instance_status=model.instance_status,
            container_name=model.container_name,
            status=model.status,
            message=model.message,
            image_name=model.image_name,
            image_digest=model.image_digest,
            restart_count=model.restart_count,
            start_time=model.start_time,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a ServiceContainer.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the ServiceContainer object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> ServiceContainerModel:
        """Create an instance of ServiceContainer from a dict.

        This method constructs a ServiceContainer object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        ServiceContainer
            A ServiceContainer object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(ServiceContainer.from_dict(obj))


ServiceContainer._model_class = ServiceContainerModel



import os
import time

from contextlib import suppress

import pytest as pytest

from snowflake.core import Clone, PointOfTimeOffset
from snowflake.core.exceptions import APIError, NotFoundError
from snowflake.core.schema import Schema
from snowflake.core.table import Table, TableColumn
from tests.utils import random_string


CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def test_create_from_model_create_mode(tables):
    table_name = random_string(10, "test_table_ INTEGRATION_")
    table_name_case_sensitive = '"' + table_name + '"'
    column_name_case_sensitive = '"cc2"'
    try:
        created_handle = tables.create(
            Table(
                name=table_name_case_sensitive,
                columns=[
                    TableColumn(name="c1", datatype="varchar"),
                    TableColumn(name=column_name_case_sensitive, datatype="varchar"),
                ],
            ),
            mode="errorifexists",
        )
        assert created_handle.name == table_name_case_sensitive
    finally:
        with suppress(Exception):
            created_handle.drop()

    # wrong name
    table_name = random_string(10, '"test_table_INTEGRATION_"')
    with pytest.raises(APIError):
        created_handle = tables.create(
            Table(name=table_name, columns=[TableColumn(name="c1", datatype="varchar")]), mode="errorifexists"
        )


@pytest.mark.skip_notebook
@pytest.mark.skip_storedproc
def test_create_using_template(session, tables):
    session.sql("create or replace temp stage table_test_stage").collect()
    table_name = random_string(10, "test_table_")
    try:
        session.sql("CREATE or replace temp FILE FORMAT table_test_csv_format TYPE = csv parse_header=true").collect()
        session.file.put(CURRENT_DIR + "/../../resources/testCSVheader.csv", "@table_test_stage", auto_compress=False)
        handle = tables.create(
            table_name,
            template="select array_agg(object_construct(*)) "
            "from table(infer_schema(location=>'@table_test_stage', "
            "file_format=>'table_test_csv_format', "
            "files=>'testCSVheader.csv'))",
        )
        table = handle.fetch()
        assert table.columns[0].name == "id"
        assert table.columns[1].name == "name"
        assert table.columns[2].name == "rating"
        assert table.rows == 0

    finally:
        with suppress(Exception):
            session.sql("drop file format if exists table_test_csv_file_format")
        with suppress(Exception):
            session.sql("drop stage if exists @table_test_stage").collect()
        with suppress(Exception):
            tables[table_name].drop()


@pytest.mark.parametrize("create_mode", ["errorifexists", "orreplace", "ifnotexists"])
def test_create_like(tables, table_handle, create_mode):
    table_name = random_string(10, "test_table_")
    created_handle = tables.create(table_name, like_table=f"{table_handle.name}", copy_grants=True, mode=create_mode)
    try:
        assert created_handle.name == table_name
    finally:
        with suppress(Exception):
            created_handle.drop()


def test_create_clone(tables, table_handle):
    table_name = random_string(10, "test_table_")
    created_handle = tables.create(
        table_name, clone_table=f"{table_handle.name}", copy_grants=True, mode="errorifexists"
    )
    try:
        assert created_handle.name == table_name
        assert created_handle.fetch().rows == 2
    finally:
        with suppress(Exception):
            created_handle.drop()

    time.sleep(1)
    table_name = random_string(10, "test_table_")
    created_handle = tables.create(
        table_name,
        clone_table=Clone(
            source=f"{table_handle.name}", point_of_time=PointOfTimeOffset(reference="before", when="-1")
        ),
        copy_grants=True,
        mode="errorifexists",
    )
    try:
        assert created_handle.name == table_name
        assert created_handle.fetch().rows == 2
    finally:
        with suppress(Exception):
            created_handle.drop()

    table_name = random_string(10, "test_table_")
    with pytest.raises(NotFoundError):
        tables.create(table_name, clone_table="non_existant_name", copy_grants=True, mode="errorifexists")


@pytest.mark.parametrize("create_mode", ["errorifexists", "orreplace", "ifnotexists"])
def test_create_as_select(tables, table_handle, create_mode):
    table_name = random_string(10, "test_table_")
    copy_grants = True if create_mode == "orreplace" else False
    created_handle = tables.create(
        table_name,
        as_select=f"select * from {table_handle.database.name}.{table_handle.schema.name}.{table_handle.name}"
        " where c1 = 1",
        copy_grants=copy_grants,
        mode=create_mode,
    )
    try:
        table = created_handle.fetch()
        assert created_handle.name == table_name
        assert table.rows == 1
    finally:
        with suppress(Exception):
            created_handle.drop()


def test_create_clone_across_schemas(table_handle, temp_schema):
    # clone <current_schema>.table to temp_schema.table_name
    table_name = random_string(10, "test_clone_table_across_schema_")

    created_handle = temp_schema.tables.create(
        Table(name=table_name), clone_table=f"{table_handle.schema.name}.{table_handle.name}"
    )

    assert created_handle.fetch().rows == 2
    assert created_handle.fetch().schema_name.upper() == temp_schema.name.upper()
    assert created_handle.name.upper() in [i.name.upper() for i in temp_schema.tables.iter()]


def test_create_clone_across_database(table_handle, temp_db):
    # clone <current_db>.<current_schema>.table to temp_db.created_schema.table_name
    schema_name = random_string(10, "test_create_clone_across_schema_")
    created_schema = temp_db.schemas.create(Schema(name=schema_name))
    table_name = random_string(10, "test_table_clone_across_database_")

    try:
        created_handle = created_schema.tables.create(
            Table(name=table_name),
            clone_table=f"{table_handle.database.name}.{table_handle.schema.name}.{table_handle.name}",
        )

        assert created_handle.fetch().rows == 2
        assert created_handle.fetch().schema_name.upper() == schema_name.upper()
        assert created_handle.fetch().database_name.upper() == temp_db.name.upper()
        assert created_handle.name.upper() in [i.name.upper() for i in created_schema.tables.iter()]
        assert created_schema.name.upper() in [i.name.upper() for i in temp_db.schemas.iter()]
    finally:
        created_schema.drop()



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import pytest

from snowflake.core.exceptions import NotFoundError
from tests.integ.managed_account.utils import (
    assert_managed_account_existence,
    cleanup_managed_accounts_completely,
    create_managed_account,
)

from ..utils import backup_role


pytestmark = [pytest.mark.internal_only, pytest.mark.skip_notebook]


def test_drop_managed_account(temp_customer_account_managed_accounts, sf_cursor, temp_customer_account_cursor):
    with backup_role(temp_customer_account_cursor):
        try:
            temp_customer_account_cursor.execute("use role accountadmin;")
            # test happy path
            managed_account = create_managed_account(
                temp_customer_account_managed_accounts, sf_cursor, temp_customer_account_cursor
            )

            # managed_account deleted
            temp_customer_account_managed_accounts[managed_account.name].drop()
            assert_managed_account_existence(
                temp_customer_account_managed_accounts, managed_account.name, should_exist=False
            )

            # managed_account already deleted
            with pytest.raises(NotFoundError):
                temp_customer_account_managed_accounts[managed_account.name].drop()

            # delete managed_account that does not exist
            with pytest.raises(NotFoundError):
                temp_customer_account_managed_accounts["RANDOM"].drop()
        finally:
            # cleanup
            cleanup_managed_accounts_completely(temp_customer_account_cursor)



"""
Snowflake Catalog Integration API.

The Snowflake Catalog Integration API is a REST API that you can use to access, update, and perform certain actions on Catalog Integration resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, ClassVar, Dict, Optional, Union

from pydantic import BaseModel, ConfigDict

import snowflake.core.catalog_integration._generated.models

from snowflake.core.catalog_integration._generated.models import *  # noqa: F403


class Catalog(BaseModel):
    """A model object representing the Catalog resource.

    Constructs an object of type Catalog with the provided properties.

    Parameters
    __________
    """

    __properties = ["catalog_source"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    # JSON field name that stores the object type
    __discriminator_property_name: ClassVar[str] = "catalog_source"

    # discriminator mappings
    __discriminator_value_class_map: ClassVar[Dict[str, str]] = {
        "GLUE": "Glue",
        "OBJECT_STORE": "ObjectStore",
        "POLARIS": "Polaris",
    }

    @classmethod
    def get_discriminator_value(cls, obj: dict[str, Any]) -> Optional[str]:
        """Returns the discriminator value (object type) of the data."""
        discriminator_value = obj[cls.__discriminator_property_name]
        if discriminator_value:
            return cls.__discriminator_value_class_map.get(discriminator_value)
        else:
            return None

    __discriminator_value_to_type: ClassVar[Dict[str, str]] = {
        "Glue": "GLUE",
        "ObjectStore": "OBJECT_STORE",
        "Polaris": "POLARIS",
    }

    @classmethod
    def get_child_model_discriminator_value(cls, child_model: str) -> str:
        return cls.__discriminator_value_to_type[child_model]

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Union[Glue, ObjectStore, Polaris]:
        """Create an instance of Catalog from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Union[Glue, ObjectStore, Polaris]:
        """Create an instance of Catalog from a dict."""
        # look up the object type based on discriminator mapping
        object_type = cls.get_discriminator_value(obj)
        if object_type:
            klass = getattr(snowflake.core.catalog_integration._generated.models, object_type)
            return klass.from_dict(obj)
        else:
            raise ValueError(
                "Catalog failed to lookup discriminator value from "
                + json.dumps(obj)
                + ". Discriminator property name: "
                + cls.__discriminator_property_name
                + ", mapping: "
                + json.dumps(cls.__discriminator_value_class_map)
            )


class CatalogModel:
    def __init__(
        self,  # optional properties
    ):
        """A model object representing the Catalog resource.

        Constructs an object of type Catalog with the provided properties.

        Parameters
        __________
        """
        pass

    __properties = ["catalog_source"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Catalog()

    @classmethod
    def _from_model(cls, model) -> CatalogModel:
        return model.__class__._model_class._from_model(model)

    def to_dict(self):
        """Create a dictionary of the properties from a Catalog.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Catalog object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> Union[GlueModel, ObjectStoreModel, PolarisModel]:
        """Create an instance of Catalog from a dict.

        This method constructs a Catalog object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Catalog
            A Catalog object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Catalog.from_dict(obj))


Catalog._model_class = CatalogModel



import pytest

from snowflake.core import CreateMode
from snowflake.core.database import Database
from snowflake.core.database_role import DatabaseRole
from snowflake.core.exceptions import ConflictError, UnauthorizedError
from tests.integ.utils import backup_database_and_schema
from tests.utils import random_string


pytestmark = pytest.mark.min_sf_ver("8.39.0")


def test_clone(databases, cursor):
    try:
        with backup_database_and_schema(cursor):
            new_db_def = Database(name=random_string(3, "test_database_$12create_"), kind="TRANSIENT")
            new_db_def.comment = "database first"
            database_handle_first = databases.create(new_db_def)
            cursor.execute(f"USE DATABASE {new_db_def.name}")

            database_role_name = random_string(10, "test_database_role")
            database_role = DatabaseRole(name=database_role_name, comment="test_comment")

            database_role_handles = []
            database_handle = None

            database_roles = databases[new_db_def.name].database_roles
            try:
                # check there are no database roles yet
                assert len(list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))) == 0

                # happy path
                database_role_handles += [database_roles.create(database_role)]
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))
                assert len(database_roles_list) == 1
                assert database_roles_list[0].name == database_role_name.upper()
                assert database_roles_list[0].comment == "test_comment"

                # clone from existing role in the same database throws error
                target_database_role_name = random_string(10, "test_database_role")
                with pytest.raises(UnauthorizedError):
                    database_roles[database_role_name].clone(target_database_role_name)

                with backup_database_and_schema(cursor):
                    new_db_def = Database(name=random_string(3, "test_database_$12create_"), kind="TRANSIENT")
                    new_db_def.comment = "database first"
                    database_handle = databases.create(new_db_def)
                    cursor.execute(f"USE DATABASE {new_db_def.name}")

                    # clone when specifying the target database
                    database_roles[database_role_name].clone(target_database_role_name, target_database=new_db_def.name)
                    database_roles_list = list(
                        databases[new_db_def.name].database_roles.iter(
                            limit=1, from_name=target_database_role_name.upper()[:-1]
                        )
                    )
                    assert len(database_roles_list) == 1

                    # clone from existing role or_replace
                    database_roles[database_role_name].clone(
                        target_database_role_name, target_database=new_db_def.name, mode=CreateMode.or_replace
                    )
                    database_roles_list = list(
                        databases[new_db_def.name].database_roles.iter(
                            limit=1, from_name=target_database_role_name.upper()[:-1]
                        )
                    )
                    assert len(database_roles_list) == 1

                    # clone from existing role if_not_exists
                    # TODO(SNOW-1707547): Uncomment this when the bug is fixed
                    # database_roles[database_role_name].clone(
                    #     target_database_role_name, target_database=new_db_def.name, mode="if_not_exists"
                    # )
                    # database_roles_list = list(
                    #     databases[new_db_def.name].database_roles.iter(
                    #         limit=1, from_name=target_database_role_name.upper()[:-1]
                    #     )
                    # )
                    # assert len(database_roles_list) == 1

                    # clone from existing role error_if_exists
                    with pytest.raises(ConflictError):
                        database_roles[database_role_name].clone(
                            target_database_role_name, target_database=new_db_def.name, mode=CreateMode.error_if_exists
                        )

                # clone when specifying the target database which is different from the current database
                database_roles[database_role_name].clone(
                    target_database_role_name, target_database=new_db_def.name, mode=CreateMode.or_replace
                )
                database_roles_list = list(
                    databases[new_db_def.name].database_roles.iter(
                        limit=1, from_name=target_database_role_name.upper()[:-1]
                    )
                )
                assert len(database_roles_list) == 1

            finally:
                database_handle.drop()
                for database_role_handle in database_role_handles:
                    database_role_handle.drop()
    finally:
        database_handle_first.drop()



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#


import pytest

from snowflake.core.database import Database
from snowflake.core.exceptions import NotFoundError

from ..utils import random_string


pytestmark = pytest.mark.usefixtures("backup_database_schema")


def test_drop(databases):
    new_database = Database(name=random_string(5, "test_db_"))
    new_db = databases.create(new_database)

    # drop a database which exists
    new_db.drop()
    with pytest.raises(NotFoundError):
        new_db.fetch()

    # drop a database which does not exists
    with pytest.raises(NotFoundError):
        new_db.drop()

    # This should not error
    new_db.drop(if_exists=True)


@pytest.mark.min_sf_ver("8.39.0")
def test_drop_and_undrop(databases):
    new_database = Database(name=random_string(5, "test_db_"))
    new_db = databases.create(new_database)

    # drop a database which exists
    new_db.drop()
    with pytest.raises(NotFoundError):
        new_db.fetch()

    # undrop a database
    new_db.undrop()

    new_db.fetch()



# coding: utf-8

# flake8: noqa
"""
Snowflake Grant API.

The Snowflake Grant API is a REST API that you can use to show or manage privileges that have been provided to users and roles in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

# import models into model package
from snowflake.core.grant._generated.models.error_response import ErrorResponse
from snowflake.core.grant._generated.models.grant import Grant
from snowflake.core.grant._generated.models.success_accepted_response import SuccessAcceptedResponse
from snowflake.core.grant._generated.models.success_response import SuccessResponse

__all__ = [
    "ErrorResponse",
    "Grant",
    "SuccessAcceptedResponse",
    "SuccessResponse",
]



"""Manages Snowflake Catalog Integrations."""

from ..catalog_integration._generated.models import CatalogIntegration, Glue, OAuth, ObjectStore, Polaris, RestConfig
from ._catalog_integration import CatalogIntegrationCollection, CatalogIntegrationResource


__all__ = [
    "CatalogIntegration",
    "CatalogIntegrationResource",
    "CatalogIntegrationCollection",
    "Glue",
    "OAuth",
    "ObjectStore",
    "Polaris",
    "RestConfig",
]



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import pytest

from snowflake.core.exceptions import ConflictError
from tests.integ.managed_account.utils import cleanup_managed_accounts_completely, create_managed_account

from ..utils import backup_role


pytestmark = [pytest.mark.internal_only, pytest.mark.skip_notebook]


def test_create_managed_account(temp_customer_account_managed_accounts, sf_cursor, temp_customer_account_cursor):
    with backup_role(temp_customer_account_cursor):
        try:
            temp_customer_account_cursor.execute("use role accountadmin;")
            # test happy path
            managed_account = create_managed_account(
                temp_customer_account_managed_accounts, sf_cursor, temp_customer_account_cursor
            )

            # managed_account already exists
            create_managed_account(
                temp_customer_account_managed_accounts,
                sf_cursor,
                temp_customer_account_cursor,
                managed_account,
                ConflictError,
            )

            # managed_account deleted
            temp_customer_account_managed_accounts[managed_account.name].drop()

            # recreate the account with the same name
            create_managed_account(
                temp_customer_account_managed_accounts, sf_cursor, temp_customer_account_cursor, managed_account
            )

            # managed_account deleted
            temp_customer_account_managed_accounts[managed_account.name].drop()
        finally:
            # cleanup
            cleanup_managed_accounts_completely(temp_customer_account_cursor)



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import pytest

from tests.integ.account.utils import cleanup_accounts_completely, create_account

from ..utils import backup_role


pytestmark = [pytest.mark.internal_only, pytest.mark.skip_notebook]


def test_list_account(temp_customer_account_accounts, sf_cursor, temp_customer_account_cursor):
    with backup_role(temp_customer_account_cursor):
        try:
            temp_customer_account_cursor.execute("use role orgadmin;")
            # test happy path
            account = create_account(temp_customer_account_accounts, sf_cursor, temp_customer_account_cursor)

            temp_customer_account_accounts[account.name].drop(grace_period_in_days=4)

            # use history and the deleted account should be there
            account_names = [account.name for account in temp_customer_account_accounts.iter(history=True)]
            assert account.name.upper() in account_names

            account = create_account(temp_customer_account_accounts, sf_cursor, temp_customer_account_cursor)
            account_2 = create_account(temp_customer_account_accounts, sf_cursor, temp_customer_account_cursor)

            account_names = [account.name for account in temp_customer_account_accounts.iter()]
            assert len(account_names) == 3  # +1 for the orgadmin account from where we are executing these statements

            account_names = [account.name for account in temp_customer_account_accounts.iter(like="")]
            assert len(account_names) == 0

            account_names = [
                account.name for account in temp_customer_account_accounts.iter(like="TEST_CREATE_ACCOUNT_%")
            ]
            assert len(account_names) == 2

            # cleanup
            temp_customer_account_accounts[account.name].drop(grace_period_in_days=4)
            temp_customer_account_accounts[account_2.name].drop(grace_period_in_days=4)
        finally:
            cleanup_accounts_completely(sf_cursor)



"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr

from snowflake.core.cortex.analyst_service._generated.models.semantic_model_object import (
    SemanticModelObject,
    SemanticModelObjectModel,
)
from snowflake.core.cortex.analyst_service._generated.models.table_columns import TableColumns, TableColumnsModel


class GenerateFiltersAndMetricsSuggestionsRequest(BaseModel):
    """A model object representing the GenerateFiltersAndMetricsSuggestionsRequest resource.

    Constructs an object of type GenerateFiltersAndMetricsSuggestionsRequest with the provided properties.

    Parameters
    __________
    semantic_model : SemanticModelObject, optional

    tables : list[TableColumns], optional
        Tables and selected columns to build a basic semantic model from
    warehouse : str, optional
        Warehouse name to use for processing suggestions.
    metrics_offset : int, optional
        Pagination offset for metrics suggestions
    metrics_limit : int, optional
        Pagination limit for metrics suggestions
    facts_offset : int, optional
        Pagination offset for facts suggestions
    facts_limit : int, optional
        Pagination limit for facts suggestions
    filters_offset : int, optional
        Pagination offset for filters suggestions
    filters_limit : int, optional
        Pagination limit for filters suggestions
    experimental : str, optional
        JSON serialized string of experimental API fields (undocumented).
    """

    semantic_model: Optional[SemanticModelObject] = None

    tables: Optional[List[TableColumns]] = None

    warehouse: Optional[StrictStr] = None

    metrics_offset: Optional[StrictInt] = None

    metrics_limit: Optional[StrictInt] = None

    facts_offset: Optional[StrictInt] = None

    facts_limit: Optional[StrictInt] = None

    filters_offset: Optional[StrictInt] = None

    filters_limit: Optional[StrictInt] = None

    experimental: Optional[StrictStr] = None

    __properties = [
        "semantic_model",
        "tables",
        "warehouse",
        "metrics_offset",
        "metrics_limit",
        "facts_offset",
        "facts_limit",
        "filters_offset",
        "filters_limit",
        "experimental",
    ]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> GenerateFiltersAndMetricsSuggestionsRequest:
        """Create an instance of GenerateFiltersAndMetricsSuggestionsRequest from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of semantic_model
        if self.semantic_model:
            _dict["semantic_model"] = self.semantic_model.to_dict()

        # override the default output from pydantic by calling `to_dict()` of each item in tables (list)
        _items = []
        if self.tables:
            for _item in self.tables:
                if _item:
                    _items.append(_item.to_dict())
            _dict["tables"] = _items

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> GenerateFiltersAndMetricsSuggestionsRequest:
        """Create an instance of GenerateFiltersAndMetricsSuggestionsRequest from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return GenerateFiltersAndMetricsSuggestionsRequest.model_validate(obj)

        _obj = GenerateFiltersAndMetricsSuggestionsRequest.model_validate(
            {
                "semantic_model": SemanticModelObject.from_dict(obj.get("semantic_model"))
                if obj.get("semantic_model") is not None
                else None,
                "tables": [TableColumns.from_dict(_item) for _item in obj.get("tables")]
                if obj.get("tables") is not None
                else None,
                "warehouse": obj.get("warehouse"),
                "metrics_offset": obj.get("metrics_offset"),
                "metrics_limit": obj.get("metrics_limit"),
                "facts_offset": obj.get("facts_offset"),
                "facts_limit": obj.get("facts_limit"),
                "filters_offset": obj.get("filters_offset"),
                "filters_limit": obj.get("filters_limit"),
                "experimental": obj.get("experimental"),
            }
        )

        return _obj


class GenerateFiltersAndMetricsSuggestionsRequestModel:
    def __init__(
        self,  # optional properties
        semantic_model: Optional[SemanticModelObject] = None,
        tables: Optional[list[TableColumns]] = None,
        warehouse: Optional[str] = None,
        metrics_offset: Optional[int] = None,
        metrics_limit: Optional[int] = None,
        facts_offset: Optional[int] = None,
        facts_limit: Optional[int] = None,
        filters_offset: Optional[int] = None,
        filters_limit: Optional[int] = None,
        experimental: Optional[str] = None,
    ):
        """A model object representing the GenerateFiltersAndMetricsSuggestionsRequest resource.

        Constructs an object of type GenerateFiltersAndMetricsSuggestionsRequest with the provided properties.

        Parameters
        __________
        semantic_model : SemanticModelObject, optional

        tables : list[TableColumns], optional
            Tables and selected columns to build a basic semantic model from
        warehouse : str, optional
            Warehouse name to use for processing suggestions.
        metrics_offset : int, optional
            Pagination offset for metrics suggestions
        metrics_limit : int, optional
            Pagination limit for metrics suggestions
        facts_offset : int, optional
            Pagination offset for facts suggestions
        facts_limit : int, optional
            Pagination limit for facts suggestions
        filters_offset : int, optional
            Pagination offset for filters suggestions
        filters_limit : int, optional
            Pagination limit for filters suggestions
        experimental : str, optional
            JSON serialized string of experimental API fields (undocumented).
        """
        self.semantic_model = semantic_model
        self.tables = tables
        self.warehouse = warehouse
        self.metrics_offset = metrics_offset
        self.metrics_limit = metrics_limit
        self.facts_offset = facts_offset
        self.facts_limit = facts_limit
        self.filters_offset = filters_offset
        self.filters_limit = filters_limit
        self.experimental = experimental

    __properties = [
        "semantic_model",
        "tables",
        "warehouse",
        "metrics_offset",
        "metrics_limit",
        "facts_offset",
        "facts_limit",
        "filters_offset",
        "filters_limit",
        "experimental",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return GenerateFiltersAndMetricsSuggestionsRequest(
            semantic_model=self.semantic_model._to_model() if self.semantic_model is not None else None,
            tables=[x._to_model() for x in self.tables] if self.tables is not None else None,
            warehouse=self.warehouse,
            metrics_offset=self.metrics_offset,
            metrics_limit=self.metrics_limit,
            facts_offset=self.facts_offset,
            facts_limit=self.facts_limit,
            filters_offset=self.filters_offset,
            filters_limit=self.filters_limit,
            experimental=self.experimental,
        )

    @classmethod
    def _from_model(cls, model) -> GenerateFiltersAndMetricsSuggestionsRequestModel:
        return GenerateFiltersAndMetricsSuggestionsRequestModel(
            semantic_model=SemanticModelObjectModel._from_model(model.semantic_model) if model.semantic_model else None,
            tables=[TableColumnsModel._from_model(x) for x in model.tables] if model.tables else None,
            warehouse=model.warehouse,
            metrics_offset=model.metrics_offset,
            metrics_limit=model.metrics_limit,
            facts_offset=model.facts_offset,
            facts_limit=model.facts_limit,
            filters_offset=model.filters_offset,
            filters_limit=model.filters_limit,
            experimental=model.experimental,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a GenerateFiltersAndMetricsSuggestionsRequest.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the GenerateFiltersAndMetricsSuggestionsRequest object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> GenerateFiltersAndMetricsSuggestionsRequestModel:
        """Create an instance of GenerateFiltersAndMetricsSuggestionsRequest from a dict.

        This method constructs a GenerateFiltersAndMetricsSuggestionsRequest object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        GenerateFiltersAndMetricsSuggestionsRequest
            A GenerateFiltersAndMetricsSuggestionsRequest object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(GenerateFiltersAndMetricsSuggestionsRequest.from_dict(obj))


GenerateFiltersAndMetricsSuggestionsRequest._model_class = GenerateFiltersAndMetricsSuggestionsRequestModel



# coding: utf-8

# flake8: noqa
"""
Cortex Inference API.

OpenAPI 3.0 specification for the Cortex REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

# import models into model package
from snowflake.core.cortex.inference_service._generated.models.anthropic_output import AnthropicOutput
from snowflake.core.cortex.inference_service._generated.models.anthropic_thinking import AnthropicThinking
from snowflake.core.cortex.inference_service._generated.models.cache_control import CacheControl
from snowflake.core.cortex.inference_service._generated.models.complete_request import CompleteRequest
from snowflake.core.cortex.inference_service._generated.models.complete_request_anthropic import (
    CompleteRequestAnthropic,
)
from snowflake.core.cortex.inference_service._generated.models.complete_request_anthropic_thinking import (
    CompleteRequestAnthropicThinking,
)
from snowflake.core.cortex.inference_service._generated.models.complete_request_messages_inner import (
    CompleteRequestMessagesInner,
)
from snowflake.core.cortex.inference_service._generated.models.complete_request_openai import CompleteRequestOpenai
from snowflake.core.cortex.inference_service._generated.models.complete_request_openai_reasoning import (
    CompleteRequestOpenaiReasoning,
)
from snowflake.core.cortex.inference_service._generated.models.complete_request_response_format import (
    CompleteRequestResponseFormat,
)
from snowflake.core.cortex.inference_service._generated.models.error_response import ErrorResponse
from snowflake.core.cortex.inference_service._generated.models.get_models_request import GetModelsRequest
from snowflake.core.cortex.inference_service._generated.models.get_models_response import GetModelsResponse
from snowflake.core.cortex.inference_service._generated.models.guardrails_config import GuardrailsConfig
from snowflake.core.cortex.inference_service._generated.models.image import Image
from snowflake.core.cortex.inference_service._generated.models.image_details import ImageDetails
from snowflake.core.cortex.inference_service._generated.models.json_content import JSONContent
from snowflake.core.cortex.inference_service._generated.models.non_streaming_complete_response import (
    NonStreamingCompleteResponse,
)
from snowflake.core.cortex.inference_service._generated.models.non_streaming_complete_response_choices_inner import (
    NonStreamingCompleteResponseChoicesInner,
)
from snowflake.core.cortex.inference_service._generated.models.non_streaming_complete_response_choices_inner_message import (
    NonStreamingCompleteResponseChoicesInnerMessage,
)
from snowflake.core.cortex.inference_service._generated.models.open_ai_output import OpenAIOutput
from snowflake.core.cortex.inference_service._generated.models.open_ai_reasoning import OpenAIReasoning
from snowflake.core.cortex.inference_service._generated.models.streaming_anthropic_output import (
    StreamingAnthropicOutput,
)
from snowflake.core.cortex.inference_service._generated.models.streaming_anthropic_thinking import (
    StreamingAnthropicThinking,
)
from snowflake.core.cortex.inference_service._generated.models.streaming_complete_response_data_event import (
    StreamingCompleteResponseDataEvent,
)
from snowflake.core.cortex.inference_service._generated.models.streaming_complete_response_data_event_choices_inner import (
    StreamingCompleteResponseDataEventChoicesInner,
)
from snowflake.core.cortex.inference_service._generated.models.streaming_open_ai_output import StreamingOpenAIOutput
from snowflake.core.cortex.inference_service._generated.models.streaming_open_ai_reasoning import (
    StreamingOpenAIReasoning,
)
from snowflake.core.cortex.inference_service._generated.models.streaming_text_content import StreamingTextContent
from snowflake.core.cortex.inference_service._generated.models.streaming_tool_use import StreamingToolUse
from snowflake.core.cortex.inference_service._generated.models.text_content import TextContent
from snowflake.core.cortex.inference_service._generated.models.tool import Tool
from snowflake.core.cortex.inference_service._generated.models.tool_choice import ToolChoice
from snowflake.core.cortex.inference_service._generated.models.tool_results import ToolResults
from snowflake.core.cortex.inference_service._generated.models.tool_results_tool_results import ToolResultsToolResults
from snowflake.core.cortex.inference_service._generated.models.tool_tool_spec import ToolToolSpec
from snowflake.core.cortex.inference_service._generated.models.tool_tool_spec_input_schema import (
    ToolToolSpecInputSchema,
)
from snowflake.core.cortex.inference_service._generated.models.tool_use import ToolUse
from snowflake.core.cortex.inference_service._generated.models.tool_use_tool_use import ToolUseToolUse
from snowflake.core.cortex.inference_service._generated.models.usage import Usage

__all__ = [
    "AnthropicOutput",
    "AnthropicThinking",
    "CacheControl",
    "CompleteRequest",
    "CompleteRequestAnthropic",
    "CompleteRequestAnthropicThinking",
    "CompleteRequestMessagesInner",
    "CompleteRequestOpenai",
    "CompleteRequestOpenaiReasoning",
    "CompleteRequestResponseFormat",
    "ErrorResponse",
    "GetModelsRequest",
    "GetModelsResponse",
    "GuardrailsConfig",
    "Image",
    "ImageDetails",
    "JSONContent",
    "NonStreamingCompleteResponse",
    "NonStreamingCompleteResponseChoicesInner",
    "NonStreamingCompleteResponseChoicesInnerMessage",
    "OpenAIOutput",
    "OpenAIReasoning",
    "StreamingAnthropicOutput",
    "StreamingAnthropicThinking",
    "StreamingCompleteResponseDataEvent",
    "StreamingCompleteResponseDataEventChoicesInner",
    "StreamingOpenAIOutput",
    "StreamingOpenAIReasoning",
    "StreamingTextContent",
    "StreamingToolUse",
    "TextContent",
    "Tool",
    "ToolChoice",
    "ToolResults",
    "ToolResultsToolResults",
    "ToolToolSpec",
    "ToolToolSpecInputSchema",
    "ToolUse",
    "ToolUseToolUse",
    "Usage",
]
